import multiprocessing
import threading
import time
from concurrent.futures import ThreadPoolExecutor

cpu_count = multiprocessing.cpu_count()


def concurrent_exec_functions(func_list, max_workers=None):
    if len(func_list) == 0:
        return []

    if len(func_list) == 1:
        try:
            return [func_list[0]()]
        except Exception as e:
            raise e

    if max_workers is not None:
        # work在[len(func_list), cup_count-1]中选择最小一个
        max_workers = min(cpu_count - 1, len(func_list))

    results = []
    with ThreadPoolExecutor(max_workers) as executor:
        futures = [executor.submit(func) for func in func_list]
        for future in futures:
            try:
                result = future.result()
                results.append(result)
            except Exception as e:
                results.append(None)  # 如果函数执行出现异常，对应位置存None
                raise e
    return results


def concurrent_exec_one_func(func, args_list, max_workers=None):
    func_list = []
    for arg in args_list:
        def inner_func(arg1=arg):
            return func(arg1)

        func_list.append(inner_func)
    return concurrent_exec_functions(func_list, max_workers=max_workers)


def concurrent_exec_multi_func_one_arg(func_list, args, max_workers=None):
    fs = []
    for func in func_list:
        def inner_func(_f=func):
            return _f(args)

        fs.append(inner_func)
    return concurrent_exec_functions(fs, max_workers=max_workers)


def timing_thread(user_func, sec):
    def _f():
        while True:
            user_func()
            time.sleep(sec)

    th = threading.Thread(target=_f)
    th.daemon = True
    th.start()


def async_thread(user_func):
    th = threading.Thread(target=user_func)
    th.daemon = True
    th.start()


class SharedVal:
    def __init__(self, val=None):
        self.val = val
        self.lock = threading.RLock()

    def get(self, dft_val=None):
        with self.lock:
            if self.val is None:
                return dft_val
            return self.val

    def set(self, val):
        with self.lock:
            self.val = val


if __name__ == '__main__':
    def print_ii(i):
        print(i * i)
        # print(j + j)


    def print_ii2(i):
        print(i + i)


    args = [(1), (2)]
    concurrent_exec_multi_func_one_arg([print_ii, print_ii2], 1)



import os
import time

from auto_easy.base.find_pic import PicFactory
from auto_easy.base.find_pic.find_pic import find_pics_v2
from auto_easy.base.find_pic.model import MPicDetV2, PicDetConf, PicV2
from auto_easy.base.windows import Window
from auto_easy.constant import TestPicDir
from auto_easy.utils import Timeout


class PicLoopDetConf:
    def __init__(self):
        self.to = 1
        self.sleep = 0.1
        self.break_func = None
        self.loop_obj_func = None


class WinFindPic(Window, PicFactory):
    def __init__(self, window_id=None, pic_dir=None):
        Window.__init__(self, window_id=window_id)
        PicFactory.__init__(self, pic_dir=pic_dir)

    def _cvt_pic(self, pic):
        # 字符串,可能是具体文件路径或者图片名(指定目录下的)
        if isinstance(pic, str) and not os.path.isfile(pic):
            return self.get_pic(pic)  # 从图片工厂获取图片
        return PicV2.new_auto(pic)

    def _cvt_pics(self, pics) -> list[PicV2]:
        if not isinstance(pics, list):
            pics = [pics]
        ans = []
        for pic in pics:
            if isinstance(pic, str) and not os.path.isfile(pic):
                ans.extend(self.get_pics(pic))  # 实现一个图片名对应多张图片的逻辑
            else:
                ans.append(PicV2.new_auto(pic))
        return ans

    def debug_find_pics(self, src, pics, pic_det_conf: PicDetConf = None) -> MPicDetV2:
        screen_img = self.capture_window()
        source = self._cvt_pic(screen_img)
        searches = self._cvt_pics(pics)
        mdet = find_pics_v2(source, searches, pic_det_conf)
        mdet.merge_same_name(dedup=True)  # 相同图片名称的检测结果合并
        return mdet

    # @timeit_decorator
    def raw_find_pics(self, pics, pic_det_conf: PicDetConf) -> MPicDetV2:
        # 支持同个图片名对应多张图片
        screen_img = self.capture_window()
        source = self._cvt_pic(screen_img)
        searches = self._cvt_pics(pics)
        mdet = find_pics_v2(source, searches, pic_det_conf)
        mdet.merge_same_name(dedup=True)  # 相同图片名称的检测结果合并
        return mdet

    def find_pics(self, pics, pic_det_conf: PicDetConf = None) -> MPicDetV2:
        return self.raw_find_pics(pics, pic_det_conf)

    def find_pics_simple(self, pics, box=None, sim=None, rgb=None, scale=None, multi_match=None):
        conf = PicDetConf()
        if box is not None:
            conf.box = box
        if sim is not None:
            conf.sim = sim
        if rgb is not None:
            conf.rgb = rgb
        if scale is not None:
            conf.cur_scale = scale
        if multi_match is not None:
            conf.multi_match = multi_match
        return self.raw_find_pics(pics, conf)

    def exists_pics(self, pics) -> bool:
        pics = self._cvt_pics(pics)
        mdet = self.raw_find_pics(pics, pic_det_conf=None, pics_det_conf=None)
        return mdet.check(includes=pics)

    def loop_find_pics(self, pics, to=1, sleep=None, min_det_num=-1, det_conf: PicDetConf = None) -> MPicDetV2:
        if to <= 0:
            sleep = 0
        if sleep is None:
            sleep = to / 3  # 暂定按检测三次来算, sleep在[0.1,1]之间
            sleep = min(sleep, 2)  # 最长sleep时间1秒
            sleep = max(sleep, 0.2)  # 最短sleep时间
            # logger.debug('sleep %.2f', sleep)

        searches = self._cvt_pics(pics)
        to = Timeout(to)
        to_searches = searches
        while to.not_timeout():
            mdet = self.raw_find_pics(to_searches, det_conf)

            exists_name = mdet.get_output_exists_names()
            to_searches = [search for search in searches if search.name not in exists_name]
            if len(to_searches) == 0:
                break

            if min_det_num > 0 and len(exists_name) >= min_det_num:
                break
            time.sleep(sleep)
        return mdet

    def loop_find_pics_not_exists(self, pics, to=1, sleep=None, det_conf: PicDetConf = None) -> MPicDetV2:
        if to > 0 and sleep is None:
            sleep = to / 3  # 暂定按检测三次来算, sleep在[0.2,1]之间
            sleep = min(sleep, 1)  # 最长sleep时间1秒
            sleep = max(sleep, 0.2)  # 最短sleep时间

        searches = self._cvt_pics(pics)
        to = Timeout(to)
        mdet = MPicDetV2()
        to_searches = searches
        while to.not_timeout():
            print(to_searches)
            mdet = self.raw_find_pics(to_searches, det_conf)
            exists_name = mdet.get_output_exists_names()
            to_searches = [search for search in to_searches if search.name in exists_name]
            if len(to_searches) == 0:
                return mdet
            time.sleep(sleep)
        return mdet


if __name__ == '__main__':
    win = WinFindPic('Phone-VB')
    win.pic_dir = TestPicDir
    start = time.time()
    mdet = win.loop_find_pics('core/test_1', to=10, sleep=1)
    # mdet = win.find_pics('core/test_1')
    print(mdet)
    print(time.time() - start)

    # exists = win.exists_pics('core/test_1')
    # print(exists)



import PIL
import cv2
import numpy as np
from PIL import ImageDraw, ImageFont

from auto_easy.base.image.cvt import img_2_pil, img_2_ndarray_rgb
from auto_easy.constant import get_test_pic
from auto_easy.models import Box
from auto_easy.utils import must_get_file, cache_with_custom_time


@cache_with_custom_time
def _get_cn_font(font_size):
    font_path = must_get_file('NotoSansSC-VariableFont_wght.ttf',
                              download_url='https://github.com/Jefung/auto_easy/raw/refs/heads/main/statics/NotoSansSC-VariableFont_wght.ttf')
    return ImageFont.truetype(font_path, font_size)


def draw_rectangles(img, boxes: list[Box] = None, title='', use_name=True, show=False, front_size=9):
    """
    在图片上绘制矩形框。

    参数:
        input_image: 可以是图像文件路径，PIL Image对象或NumPy ndarray。
        boxes: 矩形框列表，每个矩形由四个整数定义 (x1, y1, x2, y2)。
    """
    image = img_2_pil(img)

    # 创建一个画布(ImageDraw对象)
    draw = ImageDraw.Draw(image)

    font = _get_cn_font(front_size)  # 可能需要调整字体大小

    # 绘制每一个矩形
    boxes = boxes if boxes is not None else []
    for box in boxes:
        draw.rectangle([box.x1, box.y1, box.x2, box.y2], outline="red", width=1)
        text_position = (box.x1, box.y2 + 5)
        # 绘制文本
        if use_name and box.name:
            draw.text(text_position, box.name, fill="red", font=font)
    if show:
        image.show(title=title)
    return image  # 返回包含矩形的图像对象，用于进一步处理或保存


def show_image(img, name='img'):
    if isinstance(img, PIL.Image.Image):
        img.show(title=name)
        return
    if isinstance(img, np.ndarray):
        cv2.imshow(name, img)
        cv2.waitKey(0)
        cv2.destroyAllWindows()
        return
    raise Exception('show_image only support np.ndarray or Image.Image')


def show_image_table(images, column_titles):
    num_rows = len(images)  # 行数dd
    num_cols = len(images[0])  # 列数
    # 计算最大宽度和高度
    max_height = 0
    max_width = 0

    for row in images:
        for img in row:
            if img.shape[0] > max_height:
                max_height = img.shape[0]
            if img.shape[1] > max_width:
                max_width = img.shape[1]

    # 创建填充后的图像列表
    padded_images = []

    for row in images:
        padded_row = []
        for img in row:
            h, w = img.shape[:2]
            top = (max_height - h) // 2
            bottom = max_height - h - top
            left = (max_width - w) // 2
            right = max_width - w - left

            # 填充图像（使用黑色填充）
            padded_img = cv2.copyMakeBorder(img, top, bottom, left, right, cv2.BORDER_CONSTANT, value=[0, 0, 0])
            padded_row.append(padded_img)
        padded_images.append(padded_row)

    if len(padded_images) == 0:
        return
    # 将填充后的图像拼接
    combined_rows = []

    for row in padded_images:
        combined_row = cv2.hconcat(row)
        combined_rows.append(combined_row)

    # 最后将每行组合在一起
    final_combined_image = cv2.vconcat(combined_rows)

    # 在图像顶部添加标题行
    title_height = 40  # 标题行的高度
    title_image = np.zeros((title_height, final_combined_image.shape[1], 3), dtype=np.uint8)  # 创建黑色背景

    # 添加列标题
    for idx, title in enumerate(column_titles):
        # 计算每个列标题的X坐标
        text_x = int((final_combined_image.shape[1] / num_cols) * idx) + 10  # +10为文本间距
        cv2.putText(title_image, title, (text_x, int(title_height / 2)), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (255, 255, 255),
                    1, cv2.LINE_AA)

    # 拼接标题行和最终图像
    # 确保 title_image 和 final_combined_image 的列数一致
    final_with_titles = cv2.vconcat([final_combined_image])

    # 显示最终图像
    cv2.imshow('Combined Image with Titles', final_with_titles)
    cv2.waitKey(0)
    cv2.destroyAllWindows()


def show_multi_image(images):
    images = [img_2_ndarray_rgb(img) for img in images]
    # 找到最大宽度和高度
    max_height = max(img.shape[0] for img in images)
    max_width = max(img.shape[1] for img in images)

    # 创建填充后的图像列表
    padded_images = []

    for img in images:
        # 计算填充的边距
        height, width = img.shape[:2]
        top = (max_height - height) // 2
        bottom = max_height - height - top
        left = (max_width - width) // 2
        right = max_width - width - left

        # 填充图像（使用黑色填充，也可以选择其他颜色）
        padded_img = cv2.copyMakeBorder(img, top, bottom, left, right, cv2.BORDER_CONSTANT, value=[0, 0, 0])
        padded_images.append(padded_img)

    # 拼接所有填充后的图片（水平拼接）
    combined_image = cv2.hconcat(padded_images)

    # 显示拼接后的图像
    cv2.imshow('Combined Image', combined_image)
    cv2.waitKey(0)
    cv2.destroyAllWindows()
    # 拼接所有图片（水平拼接）
    # combined_image = cv2.hconcat(images)
    #
    # # 显示拼接后的图像
    # cv2.imshow('Combined Image', combined_image)
    # cv2.waitKey(0)
    # cv2.destroyAllWindows()


if __name__ == '__main__':
    draw_rectangles(get_test_pic('skill_box.bmp'), [Box(100, 100, 150, 150, name='test 中文')], show=True)



class Ctx:
    def __init__(self):
        pass


class DepInfo:
    def __init__(self, pic_includes=None, pic_excludes=None, ai_includes=None, ai_excludes=None):
        self.pic_includes = pic_includes if pic_includes is not None else []
        self.pic_excludes = pic_excludes if pic_excludes is not None else []
        self.ai_includes = ai_includes if ai_includes is not None else []
        self.ai_excludes = ai_excludes if ai_excludes is not None else []

    def merge_other(self, other):
        from auto_easy.utils import ab_union
        self.pic_includes = ab_union(self.pic_includes, other.pic_includes)
        self.pic_excludes = ab_union(self.pic_excludes, other.pic_excludes)
        self.ai_includes = ab_union(self.ai_includes, other.ai_includes)
        self.ai_excludes = ab_union(self.ai_excludes, other.ai_excludes)

    def merge_others(self, others):
        assert isinstance(others, list)
        for other in others:
            self.merge_other(other)



import numpy as np

from auto_easy.base.ai import AIModelBase, ModelMgrV2
from auto_easy.base.ai import OCR
from auto_easy.base.ai.ai_yolo import YoloObjs
from auto_easy.base.ai.superres.ai_supper_res import SuperRes
from auto_easy.base.windows import Window
from auto_easy.models import Box


class WinSupperRes(Window, ModelMgrV2):
    def __init__(self, window_id, models: list[AIModelBase], rpc_server_port=8765):
        Window.__init__(self, window_id=window_id)
        ModelMgrV2.__init__(self, models=models, rpc_server_port=rpc_server_port)

    def supper_res(self, img, scale_factor=3) -> np.array:
        models = self.get_models_by_cls(SuperRes)
        if len(models) == 0:
            raise Exception('No SuperRes model found')
        model: SuperRes = models[0]

        # img = self.capture(box)
        return model.predict(img, scale_factor)


class WinOCR(WinSupperRes):
    def __init__(self, window_id, models: list[AIModelBase], rpc_server_port=8765):
        WinSupperRes.__init__(self, window_id=window_id, models=models, rpc_server_port=rpc_server_port)

    def ocr(self, box: Box, scale_factor=1) -> YoloObjs:
        models = self.get_models_by_cls(OCR)
        if len(models) == 0:
            raise Exception('No yolo OCR found')
        model: OCR = models[0]

        img = self.capture(box)
        if scale_factor != 1:
            img = self.supper_res(img, scale_factor=scale_factor)

        return model.predict(img)


if __name__ == '__main__':
    supper_res = SuperRes()
    ocr = OCR()
    win = WinOCR('Phone-9a', models=[ocr, supper_res])
    ans = win.ocr(Box(264, 53, 398, 81), scale_factor=2)



import ctypes
import logging
import random
from functools import lru_cache

import PIL
import win32api
import win32con
import win32gui
import win32ui
from PIL import Image

from auto_easy.models import Box, Point
from auto_easy.utils import logger, sleep_with_ms
from auto_easy.utils.cache_util import func_cache_ignore_args


# 使用lru_cache装饰器，设置最大缓存数量（可选参数，这里设为None表示无限制）
@lru_cache(maxsize=None)
def find_windows_with_prefix(prefix='Phone'):
    # 通过win32gui.EnumWindows函数可以枚举系统中的所有顶级窗口。对于每个窗口，使用win32gui.GetWindowText函数获取窗口标题，然后检查窗口标题是否以指定的前缀开始，从而实现前缀模糊匹配。
    window_handles = []

    def callback(hwnd, extra):
        window_title = win32gui.GetWindowText(hwnd)
        if window_title.startswith(prefix):
            window_handles.append(
                {'hwnd': hwnd, 'title': window_title}
            )
        return True

    win32gui.EnumWindows(callback, None)
    sorted_ans = sorted(window_handles, key=lambda w: w['title'])
    if len(sorted_ans) == 0:
        raise Exception('No windows found, title prefix: {}'.format(prefix))
    return sorted_ans[0]['hwnd']


import time


class ClickStatus:
    def __init__(self, down=False, x=-1, y=-1):
        self.down = down
        self.x = x
        self.y = y

    def __str__(self):
        return f'{self.down}({self.x},{self.y})'


class Window:
    def __init__(self, window_id='', logger=logging.getLogger()):
        self.hwnd = None
        if window_id:
            self.hwnd = find_windows_with_prefix(window_id)
            logger.debug('init win_mgr, hwnd {}'.format(self.hwnd))
        self.logger = logger
        self.logger.setLevel(logging.INFO)
        msg = "[窗口管理器] 标题: {}, 句柄：{}, 大小: <{}>宽高{}".format(self.get_text(), self.hwnd,
                                                                        self.get_client_rect(),
                                                                        self.get_client_size())
        self.logger.info(msg=msg)
        self.prev_click_status = ClickStatus()
        self.latest_pil_img = None
        self.latest_pil_time = 0
        self.send_msg_count = 0

    def get_info(self):
        return "[窗口管理器] 标题: {}, 句柄：{}, 大小: <{}>宽高{}".format(self.get_text(), self.hwnd,
                                                                         self.get_client_rect(),
                                                                         self.get_client_size())

    @property
    def height(self):
        width, height = self.get_client_size()
        return height

    @property
    def width(self):
        width, height = self.get_client_size()
        return width

    @property
    def screen_box(self) -> Box:
        return Box(0, 0, self.width, self.height)

    def get_text(self):
        return win32gui.GetWindowText(self.hwnd)

    def get_client_size(self):
        left, top, right, down = win32gui.GetClientRect(self.hwnd)
        return right - left, down - top

    def get_client_rect(self):
        left, top, right, down = win32gui.GetClientRect(self.hwnd)
        return left, top, right, down

    def get_client_lt_in_screen(self):
        rect = win32gui.GetClientRect(self.hwnd)
        client_left, client_top = win32gui.ClientToScreen(self.hwnd, (rect[0], rect[1]))
        return client_left, client_top

    def get_window_lt_in_screen(self):

        window_rect = win32gui.GetWindowRect(self.hwnd)
        return window_rect[0], window_rect[1]

    def client_xy_to_window(self, x, y):
        abs_client_x, abs_client_y = self.get_client_lt_in_screen()
        abs_window_x, abs_window_y = self.get_window_lt_in_screen()
        return abs_client_x - abs_window_x + x, abs_client_y - abs_window_y + y

    def client_xy_to_screen(self, x, y):
        win_x, win_y = self.client_xy_to_window(x, y)
        abs_win_x, abs_win_y = self.get_window_lt_in_screen()
        return abs_win_x + win_x, abs_win_y + win_y

    def move_window(self, x1, y1, x2, y2):
        # print('move_window: x1: {}, y1: {}, x2: {}, y2: {}'.format(x1, y1, x2, y2))
        win32gui.MoveWindow(self.hwnd, x1, y1, x2 - x1, y2 - y1, True)
        win32gui.UpdateWindow(self.hwnd)
        time.sleep(1)

    def check_coor(self, x, y):
        left, top, right, down = self.get_client_rect()
        if not (left <= x <= right and top <= y <= down):
            raise Exception(f'xy(<{x},{y}>) is invalid')

    def left_click(self, x, y, press_sec=0.05, af_sleep=0, bf_sleep=0):
        time.sleep(af_sleep)

        self.check_coor(x, y)
        self.left_down(x, y)
        time.sleep(press_sec)
        self.left_up(x, y)

        time.sleep(bf_sleep)

    def left_click_in_box(self, box: Box, af_sleep=0, bf_sleep=0, scale=0.75):
        # 为了避免点击到边缘,这里先将区域缩小到scale比例
        scaled_box = box.copy_by_scale(scale)
        point = scaled_box.get_rand_point()
        # logger.debug(f"left_click_in_box: {scaled_box} -> {point}")
        self.left_click(point.x, point.y, af_sleep=af_sleep, bf_sleep=bf_sleep)

    def left_down(self, x, y):
        self.check_coor(x, y)
        self._mouse_click(True, x, y)

    def left_up(self, x=-1, y=-1):
        if x >= 0 and y >= 0:
            self.check_coor(x, y)
        self._mouse_click(False, x, y)

    def reset_left_click(self):
        if self.prev_click_status.down:
            self._mouse_click(False)

    def _mouse_click(self, down, x=-1, y=-1):
        # logger.debug('_mouse_click: down={}, x={}, y={}, prev_click_status: {}'.format(down, x, y, self.prev_click_status))
        lParam = None
        if x >= 0 and y >= 0:
            lParam = win32api.MAKELONG(x, y)
        # 抬起
        if not down:
            res = self._send_message(self.hwnd, win32con.WM_LBUTTONUP, win32con.MK_LBUTTON, lParam)
        else:
            # 之前是按压状态, 直接移动鼠标，不需要在按压
            # TODO: 可能需要增加超时失效机制
            if self.prev_click_status.down and self.prev_click_status.x > 0:
                res = self._send_message(self.hwnd, win32con.WM_MOUSEMOVE, win32con.MK_LBUTTON, lParam)
            else:
                res = self._send_message(self.hwnd, win32con.WM_LBUTTONDOWN, win32con.MK_LBUTTON, lParam)

        self.prev_click_status = ClickStatus(down, x, y)

    def mouse_move(self, x1, y1, x2, y2, steps=5, base_delay=0.001):
        dx = (x2 - x1) / steps
        dy = (y2 - y1) / steps
        self._mouse_click(False)

        # 速度曲线调整，例如使用二次函数模拟速度变化
        speed_curve = [i * (steps - i) + 1 for i in range(steps + 1)]
        total_speed = sum(speed_curve)

        for i in range(steps + 1):
            # 计算当前位置
            x = int(x1 + i * dx)
            y = int(y1 + i * dy)

            # 引入随机抖动
            jitter_x = random.randint(-2, 2)
            jitter_y = random.randint(-2, 2)
            x += jitter_x
            y += jitter_y

            # 根据速度曲线调整延迟
            current_speed = speed_curve[i] / total_speed
            current_delay = base_delay / current_speed

            # 引入随机延迟
            random_delay = random.uniform(0.8, 1.2) * current_delay

            self._mouse_click(True, x, y)
            time.sleep(random_delay)

        self._mouse_click(False)

    def debug_left_click(self, x, y):
        lParam = win32api.MAKELONG(x, y)
        self._send_message(self.hwnd, win32con.WM_LBUTTONDOWN, win32con.MK_LBUTTON, lParam)

    def mouse_wheel(self, x, y, down=True, times=100, loop_sleep=0.003):
        abs_x, abs_y = self.get_client_lt_in_screen()
        lParam = win32api.MAKELONG(abs_x + x, abs_y + y)
        wParam = win32api.MAKELONG(0, win32con.WM_MOUSEWHEEL)
        if down:
            wParam = win32api.MAKELONG(0, win32con.WM_MOUSEWHEEL * -1)

        # 发送鼠标滚轮向下滑动的消息
        i = 0
        while i < times:
            self._send_message(self.hwnd, win32con.WM_MOUSEWHEEL, wParam, lParam)
            time.sleep(loop_sleep)
            i += 1

    def wheel_move(self, down: bool, dis=100, point: Point = None, sleep_ms=10):
        if point is None:
            point = self.screen_box.get_mid_point(offset_rate=0.05)

        # logger.info('[大漠] 滑轮向{}移动{}'.format('下' if down else '上', dis))
        # 本机测试，调用80次，每次sleep 0.1，移动210的距离，平均每次调用移动2.6
        times = dis / 1.0
        self.mouse_wheel(point.x, point.y, down, times)

        sleep_with_ms(sleep_ms)

    def _send_message(self, *args, **kwargs):
        ok = win32gui.SendMessage(*args, **kwargs)
        if ok is not None and ok != 0:
            logger.debug(f'send message failed, {args}, {kwargs}')
        return ok

    def capture(self, box: Box = None, latest_lag=0):
        # 允许获取最近x秒内的图片,即图片复用
        if latest_lag > 0 and (time.time() - self.latest_pil_time) <= latest_lag:
            win_img = self.latest_pil_img
        else:
            win_img = self._capture_window()

        # 图片裁剪
        if box is not None and not box.is_empty():
            cropped_image = win_img.crop(box.tuple())
            return cropped_image
        return win_img

    # todo: 待废弃
    def capture_window(self, x1=-1, y1=-1, x2=-1, y2=-1) -> PIL.Image:
        win_img = self._capture_window()
        if x1 >= 0 and x2 >= 0 and y1 >= 0 and y2 >= 0:
            cropped_image = win_img.crop((x1, y1, x2, y2))
            return cropped_image
        return win_img

    # todo: 待废弃
    def capture_box(self, x1, y1, x2, y2) -> PIL.Image:
        win_img = self._capture_window()
        cropped_image = win_img.crop((x1, y1, x2, y2))
        return cropped_image

    @func_cache_ignore_args(0.1)
    def _capture_window(self) -> PIL.Image:
        # logger.debug('capture_window without cache')
        hwnd = self.hwnd
        x1, y1 = 0, 0
        x2, y2 = self.get_client_size()
        width = x2 - x1
        height = y2 - y1

        def _gen_img():
            # 获取窗口的设备上下文
            hdc = win32gui.GetWindowDC(hwnd)

            # 创建一个与窗口设备上下文兼容的设备上下文
            src_dc = win32ui.CreateDCFromHandle(hdc)

            # 创建一个位图对象，并将其选入内存设备上下文
            bmp = win32ui.CreateBitmap()
            bmp.CreateCompatibleBitmap(src_dc, width, height)

            # 创建一个内存设备上下文，用于存储位图
            mem_dc = src_dc.CreateCompatibleDC()
            mem_dc.SelectObject(bmp)
            # 将窗口内容复制到内存设备上下文中的位图, 参数为（目标位置的x和y坐标），（目标位图的宽度和高度），（源设备上下文），（源位置的x和y坐标），（光栅操作码）
            # 这里（源位置的x和y坐标）传入的时相对于整个窗口左上角位置， 需要计算
            target_x, target_y = self.client_xy_to_window(x1, y1)
            mem_dc.BitBlt((0, 0), (width, height), src_dc,
                          (target_x, target_y), win32con.SRCCOPY)
            # 获取位图的位数据
            bmp_info = bmp.GetInfo()
            bmp_bits = bmp.GetBitmapBits(True)
            # 使用PIL库将位图数据转换为图像对象
            img = Image.frombuffer('RGB',
                                   (bmp_info['bmWidth'], bmp_info['bmHeight']),
                                   bmp_bits, 'raw', 'BGRX', 0, 1)
            src_dc.DeleteDC()
            mem_dc.DeleteDC()
            win32gui.ReleaseDC(hwnd, hdc)
            win32gui.DeleteObject(bmp.GetHandle())
            return img

        screenshot = None
        for i in range(10):
            try:
                screenshot = _gen_img()
                break
            except Exception as e:
                logger.error('failed to generate screenshot, e:{}'.format(e))
                time.sleep(0.1)

        if screenshot is None:
            raise Exception('failed to generate screenshot')

        self.latest_pil_img = screenshot
        self.latest_pil_time = time.time()
        return screenshot

    def set_client_width_then_height(self, width, height):
        logger.debug('预期窗口 width: {}, height: {}'.format(width, height))
        # 设置当前进程为 DPI（每英寸点数，Dots Per Inch）感知状态
        ctypes.windll.user32.SetProcessDPIAware()

        client_width, client_height = self.get_client_size()
        logger.debug('当前窗口 width: {}, height: {}'.format(client_width, client_height))
        abs_win_x, abs_win_y = self.get_window_lt_in_screen()
        abs_client_rd_x, abs_client_rd_y = self.client_xy_to_screen(width, height)
        i = 0
        while i < 3:
            client_width, client_height = self.get_client_size()
            if abs(client_width - width) <= width * 0.05 and abs(client_height - height) <= height * 0.05:
                break
            i += 1
            self.move_window(abs_win_x, abs_win_y, abs_client_rd_x, abs_client_rd_y)
        logger.debug('最新窗口({}) width: {}, height: {}'.format(self.get_text(), client_width, client_height))


if __name__ == '__main__':
    pass



import time
from unittest import TestCase

from auto_easy.base.ai.model_mgr_v2 import AIModelBase, ModelMgrV2
from auto_easy.utils import async_thread


class ModelTest1(AIModelBase):
    def __init__(self):
        super().__init__(name='test1')

    def init_model(self):
        print('init model1')
        time.sleep(2)

    def predict(self, arg1):
        print('ModelTest1 input', arg1)
        return ['ModelTest1', arg1]


class TestModelMgr(TestCase):
    def test_predict(self):
        model1 = ModelTest1()
        mgr = ModelMgrV2([model1])
        async_thread(mgr.start_rpc_server)
        time.sleep(3)
        print('开始调用')
        model1_new = ModelTest1()
        mgr2 = ModelMgrV2([model1_new])
        time.sleep(1)
        ans = mgr2.predict('test1', 'input_test')
        print('resp ', ans)



import importlib.util
import os
from typing import List

from auto_easy.base.ai import ModelMgrV2, AIModelBase
from auto_easy.base.ai.ai_yolo import AIYolo
from auto_easy.base.windows import Window
from auto_easy.models import MYoloItem, YoloItem
from auto_easy.utils.cache_util import cache_with_custom_time


class WinYolo(Window, ModelMgrV2):
    def __init__(self, window_id, models: list[AIModelBase], rpc_server_port=8765, item_model_dir=''):
        Window.__init__(self, window_id=window_id)
        ModelMgrV2.__init__(self, models=models, rpc_server_port=rpc_server_port)
        self._cur_yolo_name = ''
        if item_model_dir != '' and not os.path.isdir(item_model_dir):
            raise Exception('item_model_dir must be a valid path: {}'.format(item_model_dir))
        self._item_model_dir = item_model_dir

    @property
    def cur_yolo_name(self):
        return self._cur_yolo_name

    @cur_yolo_name.setter
    def cur_yolo_name(self, name):
        yolo_models = self.get_models_by_cls(AIYolo)
        models_name = [model.name for model in yolo_models]
        if name not in models_name:
            raise Exception(f'invalid yolo model({name}), registed names: {models_name}')
        self._cur_yolo_name = name

    def yolo_predict(self, name=None) -> MYoloItem:
        yolo_models = self.get_models_by_cls(AIYolo)
        if len(yolo_models) == 0:
            raise Exception('No yolo models found')
        # 优先使用用户指定的模型,其次使用cur_yolo_name, 最后默认选择第一个
        model_name = self.cur_yolo_name if self.cur_yolo_name != '' else yolo_models[0].name
        if name is not None:
            model_name = name

        model: AIYolo = None
        for m in yolo_models:
            if m.name == model_name:
                model = m
                break
        if model is None:
            raise Exception(f'no yolo model named "{model_name}" found')

        img = self.capture()
        mdet: MYoloItem = model.predict(img)
        for i, item in enumerate(mdet.items):
            mdet.items[i] = self._auto_cvt_item(item)  # 类型转化
        return mdet

    def _auto_cvt_item(self, item: YoloItem) -> YoloItem:
        # 自动读取_item_model_dir目录下的AIItemBase类，然后调用其静态方法can_new_obj判断是否生成对象
        # 注：这里优先遍历派生类
        if self._item_model_dir == '':
            return item
        item_cls_list = find_classes_inheriting(self._item_model_dir, YoloItem)
        for cls in item_cls_list:
            if cls.can_new_obj(item.name):
                return cls(*item.params_list)
        return item


@cache_with_custom_time()
def find_classes_inheriting(dir: str, cls) -> List[type]:
    """
    在指定目录下查找继承自AIItemBase的所有类。
    :param dir: 要查找的目录路径
    :return: 继承自AIItemBase的类的列表
    """
    if not os.path.exists(dir):
        raise Exception("不存在的路径: {}".format(dir))
    if isinstance(cls, type):
        cls = cls.__name__
    if not isinstance(cls, str):
        raise Exception("find_classes_inheriting must use class or class_name")
    result_classes = []
    processed_modules = set()
    for root, dirs, files in os.walk(dir):
        for file in files:
            if file.endswith('.py') and file != '__init__.py':
                pass
                file_path = os.path.join(root, file)
                module_name = os.path.splitext(file)[0]
                spec = importlib.util.spec_from_file_location(module_name, file_path)
                module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(module)
                processed_modules.add(module_name)
                for name, obj in vars(module).items():
                    if isinstance(obj, type):
                        try:
                            if cls in [cls.__name__ for cls in obj.mro()]:
                                result_classes.append(obj)
                        except AttributeError:
                            continue
    # 因为同个文件夹下，不同文件导入的class可能是路径不同，被认为不同的类，这里通过类名统一去重
    unique_class_list = [cls for i, cls in enumerate(result_classes) if
                         cls.__name__ not in [c.__name__ for c in result_classes[:i]]]

    # 排序，将派生类排在前面
    def sort_classes(cls1, cls2):
        return len(cls1.mro()) > len(cls2.mro())

    sorted_class_list = sorted(unique_class_list, key=lambda x: tuple(sort_classes(x, y) for y in unique_class_list),
                               reverse=True)

    return sorted_class_list


if __name__ == '__main__':
    pass



import json
import socket
import time
import xmlrpc
import xmlrpc.server
from abc import abstractmethod

from auto_easy.utils import async_thread, cvt_chinese, logger, is_actual_subclass


class AIModelBase:
    def __init__(self, name, rpc_support=True, preload=True):
        self.name = name
        self.inited = False
        self.rpc_support = rpc_support
        self.model_mgr: ModelMgrV2 = None
        self.start_init = False
        if preload:
            async_thread(self._async_init)

    def _async_init(self):
        # 在这里实现异步初始化的通用逻辑
        self.start_init = True
        self.init_model()
        self.inited = True
        logger.debug("init model: {}".format(self.name))

    def wait_model_init(self):
        if not self.start_init:
            async_thread(self._async_init)
        while not self.inited:
            time.sleep(0.1)

    def predict(self, *args, **kwargs):
        # 检查模型是否已初始化，如果未初始化则等待初始化完成
        self.wait_model_init()
        return self.predict(*args, **kwargs)

    def rpc_req_encode(self, *args, **kwargs):
        return args, kwargs

    def rpc_req_decode(self, *args, **kwargs):
        return args, kwargs

    def rpc_resp_decode(self, str_resp):
        return json.loads(str_resp)

    def rpc_resp_encode(self, origin_resp):
        return json.dumps(origin_resp)

    def rpc_server_api(self, *args, **kwargs):
        args, kwargs = self.rpc_req_decode(*args, **kwargs)
        resp = self.predict(*args, **kwargs)
        return self.rpc_resp_encode(resp)

    def rpc_client_call(self, f, *args, **kwargs):
        args, kwargs = self.rpc_req_encode(*args, **kwargs)
        resp = f(*args, **kwargs)
        return self.rpc_resp_decode(resp)

    def rpc_api_name(self):
        name = '{}_{}_predict'.format(self.__class__.__name__, self.name)
        name = cvt_chinese(name)
        return name

    def mock_rpc_call(self, *args, **kwargs):
        return self.rpc_client_call(self.rpc_server_api, *args, **kwargs)

    def use_rpc(self):
        return not self.inited and self.rpc_support and self.model_mgr and self.model_mgr.rpc_client

    def rpc_call(self, *args, **kwargs):
        if not (self.model_mgr and self.model_mgr.rpc_client):
            raise RuntimeError('model_mgr not initialized')
        rpc_func = self.model_mgr.get_rpc_func(self.rpc_api_name())
        return self.rpc_client_call(rpc_func, *args, **kwargs)

    def set_model_mgr(self, model_mgr):
        self.model_mgr = model_mgr

    @abstractmethod
    def init_model(self):
        pass

    @abstractmethod
    def predict(self, *args, **kwargs):
        pass


class ModelMgrV2:
    def __init__(self, models: list[AIModelBase], rpc_server_port=8765):
        self._models = models
        self.rpc_client = None
        self.rpc_server_port = rpc_server_port
        # async_thread(self._init_rpc_client)
        self._init_rpc_client()
        for model in self._models:
            model.set_model_mgr(self)

    def predict(self, name, *args, **kwargs):
        model = self.get_model(name)
        if model is None:
            raise RuntimeError('No such model: {}'.format(name))

        # 本地模型已初始化,直接使用本地模型
        if model.inited:
            return model.predict(*args, **kwargs)

        # 远程rpc客户端已初始化, 直接使用远程客户端
        if self.rpc_client and model.rpc_support:
            return self._rpc_call(name, *args, **kwargs)

        # 使用本地模型, 内部会自动等待完成初始化
        return model.predict(*args, **kwargs)

    def get_models_by_cls(self, cls):
        return [model for model in self._models if is_actual_subclass(model, cls)]

    def get_model(self, name):
        for model in self._models:
            if model.name == name:
                return model
        return None

    def start_rpc_server(self):
        port_usable = False
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            port_usable = s.connect_ex(('127.0.0.1', self.rpc_server_port)) != 0
        if not port_usable:
            raise RuntimeError('Port {} is not available'.format(self.rpc_server_port))

        def ping():
            return 'pong'

        # # 创建服务器并注册函数
        server = xmlrpc.server.SimpleXMLRPCServer(('127.0.0.1', self.rpc_server_port))
        server.register_function(ping, 'ping')
        for model in self._models:
            server.register_function(model.rpc_server_api, model.rpc_api_name())
        # 启动服务
        server.serve_forever()

    def _init_rpc_client(self):
        try:
            server = xmlrpc.client.ServerProxy('http://127.0.0.1:{}/'.format(self.rpc_server_port))
            if server.ping() == 'pong':
                self.rpc_client = server

        except Exception as e:
            raise e

    def _rpc_call(self, name, *args, **kwargs):
        model = self.get_model(name)
        api_name = model.rpc_api_name()
        rpc_func = getattr(self.rpc_client, api_name)
        return model.rpc_client_call(rpc_func, *args, **kwargs)

    def get_rpc_func(self, api_name):
        rpc_func = getattr(self.rpc_client, api_name)
        return rpc_func



from auto_easy.core.core import CoreConf, AutoCore

conf = CoreConf()
conf.window_id = 'Phone-9a'
# conf.pic_dir = os.path.join(os.path.dirname(__file__), 'det_pic/pics')
# conf.pic_save_dir = r'C:\Users\Administrator\Desktop\自动截图'
# conf.models = [yolo_dhxbx, yolo_fsdsyc1, yolo_nshy, yolo_adgd, yolo_lp, ai_ocr, ai_supper_res]
# conf.item_model_dir = os.path.join(root_path, 'biz/models')
auto_core = AutoCore(conf)
auto_core.show()



import copy


class ConfItem:
    def __init__(self, name, value):
        self.name = name
        self.value = value


class ConfBase:
    def __init__(self):
        self._property_set_dict = {}  # key为被设置的key
        self._test_property = None

    def __setattr__(self, name, value):
        if self.is_property_attr(name):
            self._property_set_dict[name] = ConfItem(name, value)
        super().__setattr__(name, value)  # 其他属性正常设置

    def is_property_attr(self, key):
        if not hasattr(self.__class__, key):
            return False
        if isinstance(getattr(self.__class__, key), property):
            return True
        return False

    def get_property_set_dict(self):
        return self._property_set_dict

    @property
    def test_property(self):
        return self._test_property

    @test_property.setter
    def test_property(self, value):
        self._test_property = value

    def copy(self):
        obj = copy.deepcopy(self)
        obj._property_set_dict = {}
        return obj

    @staticmethod
    def new_conf_by_pry(confs):
        confs = [conf for conf in confs if conf is not None]
        if len(confs) == 0:
            return None
        base_conf = confs[0].copy()
        for conf in confs[1:]:
            for item in conf.get_property_set_dict().values():
                setattr(base_conf, item.name, item.value)
        return base_conf


# 派生类
class DerivedClass(ConfBase):
    def __init__(self):
        super().__init__()
        self._value3 = None

    @property
    def value3(self):
        return self._value3

    @value3.setter
    def value3(self, new_value):
        self._value3 = new_value



from pathlib import Path

from PIL import Image


def is_image(file_path):
    try:
        img = Image.open(file_path)  # 尝试打开图像文件  
        img.verify()  # 验证文件是否损坏  
        return True
    except (IOError, SyntaxError):  # PIL无法识别的文件类型会抛出IOError  
        return False


def get_app_dir():
    app_data_roaming = os.getenv('APPDATA')
    app_dir = os.path.join(app_data_roaming, 'auto_easy')
    if not os.path.exists(app_dir):
        os.makedirs(app_dir)
    return app_dir


def get_files(dir_path, file_prefix='', file_ext=''):
    res = []
    for root, dirs, files in os.walk(dir_path):
        for file in files:
            if file_prefix != '' and not file.startswith(file_prefix):
                continue
            if file_ext != '' and not file.endswith(file_ext):
                continue
            abs_path = os.path.join(root, file)
            res.append(abs_path)
    return res


def get_statics_dir():
    current_path = Path(__file__).resolve()
    parent_path = current_path.parents[2]
    return os.path.join(parent_path, 'statics')


import requests
import os


def download_file(url, save_path) -> str:
    """
    此函数用于从指定的 URL 下载文件并保存到指定路径。

    :param url: 要下载文件的 URL
    :param save_path: 文件保存的路径
    :return: 如果下载成功返回 True，否则返回 False
    """
    print('downloading {} -> {}'.format(url,save_path))
    try:
        # 发送 HTTP 请求，设置 stream=True 以支持流式下载
        response = requests.get(url, stream=True)
        # 检查响应状态码，确保请求成功
        response.raise_for_status()

        # 打开指定路径的文件以二进制写入模式
        with open(save_path, 'wb') as file:
            # 遍历响应内容的每个数据块
            for chunk in response.iter_content(chunk_size=8192):
                # 如果数据块不为空，则写入文件
                if chunk:
                    file.write(chunk)
        return ''
    except requests.RequestException as e:
        return f"下载文件时发生错误: {e}"
    except IOError as e:
        return f"保存文件时发生错误: {e}"


def must_get_file(filename, download_url=None):
    dirs = [
        get_statics_dir(),
        get_app_dir(),
    ]
    for d in dirs:
        file_path = os.path.join(d, filename)
        if os.path.exists(file_path):
            return file_path

    if download_url is not None:
        file_path = os.path.join(get_app_dir(), filename)
        err_msg = download_file(download_url, file_path)
        if err_msg != '':
            raise Exception(err_msg)
        return file_path

    raise Exception(f'file not found: {filename}')



import cv2
import numpy as np

from auto_easy.base.image.cvt import img_2_ndarray_rgb
from auto_easy.constant import get_test_pic


def parse_color_range(color_range):
    """
    根据用户提供的偏差参数解析上下限颜色
    :param color_range: 偏差格式字符串，例如 "F7E97D-08123D"
    :return: lower_bound, upper_bound 两个数组
    """
    # 分成基准颜色和偏差范围
    base_color, delta_color = color_range.split('-')

    # 将颜色值从 16 进制转为 RGB 三个独立数值
    base = np.array([int(base_color[i:i + 2], 16) for i in (0, 2, 4)])
    delta = np.array([int(delta_color[i:i + 2], 16) for i in (0, 2, 4)])

    # 计算颜色的上下限
    lower_bound = np.clip(base - delta, 0, 255)  # 防止低于 0
    upper_bound = np.clip(base + delta, 0, 255)  # 防止超过 255

    return lower_bound, upper_bound


# RGB 格式
def find_color(image, color_range):
    """
    在图片中寻找符合条件的颜色范围，并计算面积占比
    :param image: 输入图片路径
    :param color_range: 偏差参数，格式如 "F7E97D-08123D"
    :return: (是否存在, 占比百分比)
    """
    # 解析颜色区间
    lower_bound, upper_bound = parse_color_range(color_range)
    image = img_2_ndarray_rgb(image)

    # 根据颜色范围生成掩码
    mask = cv2.inRange(image, lower_bound, upper_bound)

    # 计算符合条件的像素数量
    matching_pixels = np.count_nonzero(mask)

    # 计算图片总像素数量
    total_pixels = image.shape[0] * image.shape[1]

    # 是否存在符合条件的颜色
    exists = matching_pixels > 0

    # 计算面积占比（百分比输出）
    area_percentage = (matching_pixels / total_pixels) * 100

    return exists, round(area_percentage, 2)


def find_best_color(image, colors_range: list[str], area_rate=0.00) -> str:
    target_color = ''
    max_rate = 0
    for color_range in colors_range:
        exists, area_percentage = find_color(image, color_range)
        if area_percentage < area_rate:
            continue
        if area_percentage >= max_rate:
            max_rate = area_percentage
            target_color = color_range
    return target_color


# 示例调用
if __name__ == "__main__":
    # skill_color_buff = '2A98B0-2A504F'
    # # skill_color_yellow = 'A38C5D-5C705E'
    # skill_beidong = '3B814E-2C3327'
    # skill_color_yellow = 'DCC151-0D174D'
    skill_color_buff = '2A98B0-2A504F'
    skill_beidong = '3B814E-2C3327'
    skill_color_yellow = 'DCC151-0D174D'
    skill_color_red = 'CE9338-263832'
    # 图片路径
    image_path = get_test_pic('debug/wujinbodong.bmp')  # 请替换为本地图片路径
    image = cv2.imread(image_path)
    # 转换图片为 RGB 格式 (OpenCV 默认是 BGR)
    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    print(find_color(image, skill_color_yellow))



import time
from collections import deque


class TimeSortedQueue:
    def __init__(self, max_size):
        """
        初始化 TimeSortedQueue 类

        :param max_size: 队列的最大长度
        """
        self.max_size = max_size
        self.queue = deque()  # 使用deque作为底层数据结构实现队列

    def put(self, data):
        """
        将数据放入队列，按照写入时间排序，若超过队列最大长度，则废弃最老的数据

        :param data: 要放入队列的任意类型数据
        """
        cur_time_ms = int(time.time() * 1000)  # 获取当前时间戳
        element = (cur_time_ms, data)  # 将时间戳和数据组成元组
        self.queue.append(element)  # 先添加到队列末尾
        self.queue = deque(sorted(self.queue, key=lambda x: x[0]))  # 按照时间戳对队列元素重新排序
        if len(self.queue) > self.max_size:
            self.queue.popleft()  # 移除最老的数据（队列头部元素）

    def get_all(self, max_lag_ms=-1, oldest_lag_ms=-1, desc=False):
        """
        返回队列中的所有元素以及对应的写入时间

        :return: 包含元素和写入时间的元组列表，按照写入时间从早到晚排序
        """
        ans = []
        cur_time_ms = int(time.time() * 1000)
        ts_list = []
        for item in list(self.queue):
            item_time_ms = item[0]
            item_val = item[1]
            if max_lag_ms == -1 or cur_time_ms - item_time_ms <= max_lag_ms:
                ans.append(item_val)
                ts_list.append(item_time_ms)

        if oldest_lag_ms >= 0 and len(ts_list) > 0 and cur_time_ms - min(ts_list) < oldest_lag_ms:
            return []

        if desc:
            return list(ans)
        return list(reversed(ans))

    def get_queue_vals(self):
        return list(reversed(list(self.queue)))

    def get(self, max_lag_ms=-1):
        if len(self.queue) == 0:
            return None
        newest = list(reversed(list(self.queue)))[0]
        newest_time_ms = newest[0]
        newest_val = newest[1]
        cur_time_ms = int(time.time() * 1000)
        if max_lag_ms <= 0 or cur_time_ms - newest_time_ms <= max_lag_ms:
            return newest_val
        return None

    def __iter__(self):
        """
        实现迭代器协议，使得类的实例可以被迭代，按写入时间顺序遍历元素和时间戳
        """
        return iter(self.queue)


if __name__ == '__main__':
    pass
    queue_obj = TimeSortedQueue(max_size=5)
    for i in range(10):
        time.sleep(0.05)
        queue_obj.put(f"element_{i}")
        # print(queue_obj.get())

    # 遍历队列元素和写入时间
    for element in queue_obj:
        print(element)

    # 获取所有元素和写入时间
    print(queue_obj.get_all())



import random
from datetime import datetime

from pypinyin import lazy_pinyin

from auto_easy.utils.time_util import Timeout


def gen_lambda_func(func, *args, **kwargs):
    def inner_func():
        return func(*args, **kwargs)

    return inner_func


count = 0


def ten_times_one_true(x=0):
    global count
    count += 1
    if count % 10 == 0:
        result = True
        count = 0
    else:
        result = False
    return result


def array_to_camelcase(arr, first=False):
    result = ""
    for s in arr:
        for index, element in enumerate(s):
            if first:
                if index == 0:
                    result += element.capitalize()
            else:
                if index == 0:
                    result += element.capitalize()
                else:
                    result += element.lower()
    return result


def cvt_chinese(s):
    py_list = lazy_pinyin(s)
    return ''.join(py_list)


import time


def loop_with_sleep(times, sleep_sec, func, logger=None):
    """
    循环调用指定函数并在每次调用后暂停指定秒数。

    参数:
    x (int): 循环的次数。
    sec (float): 每次调用函数后暂停的秒数。
    func (function): 要循环调用的函数，该函数不接受任何参数。
    """
    for i in range(int(times)):
        if logger is not None:
            logger.info("####### 开始第{}次调用 ####################".format(i + 1))
        ans = func()
        if logger is not None:
            logger.info("调用结果：{}".format(ans))
            logger.info("####### 结束第{}次调用，sleep {} ##########".format(i + 1, sleep_sec))
        time.sleep(sleep_sec)


def loop_until_true(f, to_ms=1000, sleep_ms=50, ) -> (bool, int):
    if to_ms <= 0:
        return f(), 1
    to = Timeout(to_ms / 1000)
    cnt = 0
    while to.not_timeout():
        cnt += 1
        ok = f()
        if ok:
            return True, cnt
        time.sleep(sleep_ms / 1000)
    return False, cnt


import tempfile
import os

temp_dir = tempfile.gettempdir()


def get_tmp_file(ext='jpg'):
    now = datetime.now()
    # 格式化时间字符串
    time_str = now.strftime('%Y%M%d_%H%M%S_%f')[:-3]
    return os.path.join(temp_dir, f"{time_str}.{ext}")


def set_env(k, v):
    os.environ[k] = v


def get_env(k, default=None):
    if k in os.environ:
        return os.environ[k]
    return default


def rand_int_in_time_range(min_val, max_val, time_window_seconds):
    """
    在给定的最小值和最大值之间生成随机数，并且保证在特定时间窗口内生成的随机数相同。

    参数:
    min_val (int或float): 随机数范围的最小值。
    max_val (int或float): 随机数范围的最大值。
    time_window_seconds (int): 时间窗口的秒数，在这个时间窗口内生成的随机数保持一致。

    返回:
    int或float: 生成的随机数
    """
    current_time = int(time.time())
    # 计算当前时间所在的时间窗口起始时间戳
    window_start_time = current_time - (current_time % time_window_seconds)
    random.seed(window_start_time)
    return int(random.uniform(min_val, max_val))



import time
from abc import ABC, abstractmethod
from random import uniform

from auto_easy.core.core import get_auto_core
from auto_easy.models import Ctx
from auto_easy.utils import loop_until_true, logger, cost_ms, sleep_with_rand


class Executor(ABC):
    def __init__(self, name, print_hit_false=True):
        self.name = name
        self.hit_loop_toms = 0
        self.print_hit_false = print_hit_false
        self.init()

    def init(self) -> bool:
        return True

    def reset(self, ctx: Ctx) -> bool:
        return True

    def hit(self, ctx: Ctx) -> bool:
        return True

    @abstractmethod
    def exec(self, ctx: Ctx) -> bool:
        return True

    def run(self, ctx: Ctx):
        start = time.time()

        is_hit, _ = loop_until_true(lambda: self.hit(ctx), to_ms=self.hit_loop_toms)
        if not is_hit:
            if self.print_hit_false:
                logger.debug("[执行] 前置校验失败: {}".format(self.name))
            return False

        if not self.exec(ctx):
            logger.debug("[执行] 中途执行失败: {}".format(self.name))
            return False

        logger.debug("[执行] {} 执行完成, 耗时： {}ms".format(self.name, cost_ms(start)))
        return True

    # todo: 待废弃
    def reset(self, ctx: Ctx):
        pass

    @classmethod
    def __subclasshook__(cls, subclass):
        """
        This makes the abstract method optional for the subclass to override,
        ensuring that BaseClass can still be instantiated if necessary.
        """
        if cls is Executor:
            return True
        return NotImplemented


class ExecutorDebug(Executor):
    def __init__(self, name, hit_ret=True, exec_ret=True, hit_wait=0, exec_wait=0):
        super().__init__("Debug执行器{}".format(name), print_hit_false=False)
        self.hit_ret = hit_ret
        self.exec_ret = exec_ret
        self.exec_wait = exec_wait
        self.hit_wait = hit_wait

    def hit(self, ctx: Ctx) -> bool:
        time.sleep(self.hit_wait)
        return self.hit_ret

    def exec(self, ctx: Ctx) -> bool:
        time.sleep(self.exec_wait)
        return self.exec_ret


class ExecutorPicClick(Executor):
    def __init__(self, pic_name, det_to=2, bf_sleep=0.2, af_sleep=0.2, x_offset=0, y_offset=0, hit_err_print=True):
        self.pic_name = pic_name
        self.det_to = det_to
        self.bf_sleep = bf_sleep
        self.af_sleep = af_sleep
        self.x_offset = x_offset  # 点击偏移量
        self.y_offset = y_offset  # 点击偏移量
        self.hit_err_print = hit_err_print
        super().__init__(name='图片点击({})'.format(pic_name))

    def hit(self, ctx: Ctx) -> bool:
        mdet = get_auto_core().loop_find_pics(self.pic_name, to=self.det_to)
        if not mdet.all_detected:
            if self.hit_err_print:
                pass
                # logger.debug("[图片点击] 识别失败，无法识别. {}".format(self.pic_name))
            return False
        return True

    def exec(self, ctx: Ctx) -> bool:
        mdet = get_auto_core().loop_find_pics(self.pic_name, to=self.det_to)
        if not mdet.all_detected:
            logger.error("[图片点击] 识别失败，无法识别. {}".format(self.pic_name))
            return False
        logger.debug("[图片点击] 点击图片({}), 区域: {}".format(self.pic_name, mdet.box))

        box = mdet.box
        box.move(self.x_offset, self.y_offset)
        get_auto_core().left_click_in_box(
            box,
            bf_sleep=self.bf_sleep * uniform(0.8, 1.2),
            af_sleep=self.af_sleep * uniform(0.8, 1.2),
        )
        return True


class ExecutorTryPicClick(Executor):
    def __init__(self, pic_name, det_to=0.5, bf_sleep=0.2, af_sleep=0.3):
        self.pic_name = pic_name
        self.det_to = det_to
        self.bf_sleep = bf_sleep
        self.af_sleep = af_sleep
        super().__init__(name='图片点击({})'.format(pic_name))

    def hit(self, ctx: Ctx) -> bool:
        return True

    def exec(self, ctx: Ctx) -> bool:
        mdet = get_auto_core().loop_find_pics(self.pic_name, to=self.det_to)
        if not mdet.is_detected:
            return True

        logger.info("[点击] 点击图片({}), 区域: {}".format(self.pic_name, mdet.box))
        get_auto_core().left_click_in_box(
            mdet.box,
            af_sleep=self.af_sleep * uniform(0.8, 1.2),
            bf_sleep=self.bf_sleep * uniform(0.8, 1.2),
        )
        return True


class ExecutorPicDet(Executor):
    def __init__(self, pic_name, det_to=2, af_sleep=0):
        super().__init__(name='图片检测({})'.format(pic_name))
        self.pic_name = pic_name
        self.det_to = det_to
        self.af_sleep = af_sleep

    def hit(self, ctx: Ctx) -> bool:
        det = get_auto_core().loop_find_pics(self.pic_name, to=self.det_to)
        if not det.is_detected:
            return False
        return True

    def exec(self, ctx: Ctx) -> bool:
        sleep_with_rand(self.af_sleep)
        return True


class ExecutorPicDisappear(Executor):
    def __init__(self, pic_name, det_to=0, wait_to=5):
        super().__init__(name='图片检测({})'.format(pic_name))
        self.pic_name = pic_name
        self.det_to = det_to
        self.wait_to = wait_to

    def hit(self, ctx: Ctx) -> bool:
        det = get_auto_core().loop_find_pics(self.pic_name, to=self.det_to)
        if not det.is_detected:
            return False
        return True

    def exec(self, ctx: Ctx) -> bool:
        mdet = get_auto_core().loop_find_pics_not_exists(self.pic_name, to=self.wait_to)

        if mdet.is_detected:
            return False
        return True


class ExecutorPicTFSwitch(Executor):
    def __init__(self, true_pic, false_pic, want_true, det_to=2, af_sleep=0.5):
        super().__init__(name='图片开关检测({}-{})[{}]'.format(true_pic, false_pic, want_true))
        self.want_true = want_true
        self.true_pic = true_pic
        self.false_pic = false_pic
        self.det_to = det_to
        self.af_sleep = af_sleep

    @property
    def pics_name(self):
        return [self.true_pic, self.false_pic]

    def hit(self, ctx: Ctx) -> bool:
        # 至少检测一张图片
        det = get_auto_core().loop_find_pics(self.pics_name, to=self.det_to, min_det_num=1)
        if not det.is_detected:
            logger.debug('同时不存在两张图片')
            return False
        if det.check(includes=self.pics_name):
            return False
        return True

    def exec(self, ctx: Ctx) -> bool:
        mdet = get_auto_core().loop_find_pics(self.pics_name, to=self.det_to, min_det_num=1)
        to_click_pic = ''
        if self.want_true and mdet.check(includes=[self.false_pic]):
            to_click_pic = self.false_pic

        if not self.want_true and mdet.check(includes=[self.true_pic]):
            to_click_pic = self.true_pic

        if to_click_pic == '':
            return True
        box = mdet.get(to_click_pic).box
        logger.debug(f'为了达到目标({self.want_true}), 点击图片({to_click_pic}-{box})')
        get_auto_core().left_click_in_box(
            mdet.box,
            af_sleep=self.af_sleep * uniform(0.8, 1.2),
        )
        return True



from abc import ABC, abstractmethod

import cv2
import numpy as np

from auto_easy.base.image.process import image_color_keep
from auto_easy.models import Box


class PicProcess(ABC):
    def __init__(self, name):
        self.name = name

    @abstractmethod
    def process(self, img_source_rgb: np.ndarray, img_search_rgb: np.ndarray) -> (np.ndarray, np.ndarray):
        pass


class PicResize(PicProcess):
    def __init__(self, scale=1):
        super().__init__(name='图片缩放')
        self.scale = scale

    def process(self, img_source_rgb: np.ndarray, img_search_rgb: np.ndarray) -> (np.ndarray, np.ndarray):
        img_search_rgb = cv2.resize(
            img_search_rgb,
            None,
            fx=self.scale,
            fy=self.scale,
            interpolation=cv2.INTER_CUBIC
        )
        return img_source_rgb, img_search_rgb


class PicColorKeep(PicProcess):
    def __init__(self, color=None):
        super().__init__(name='颜色保留')
        self.color = color

    def process(self, img_source_rgb: np.ndarray, img_search_rgb: np.ndarray) -> (np.ndarray, np.ndarray):
        img_source_rgb = image_color_keep(img_source_rgb, self.color)
        img_search_rgb = image_color_keep(img_search_rgb, self.color)
        return img_source_rgb, img_search_rgb


class PicColorGray(PicProcess):
    def __init__(self, ):
        super().__init__('转灰度图片')

    def process(self, img_source_rgb: np.ndarray, img_search_rgb: np.ndarray) -> (np.ndarray, np.ndarray):
        img_source_gray = cv2.cvtColor(img_source_rgb, cv2.COLOR_BGR2GRAY)
        img_search_rgb = cv2.cvtColor(img_search_rgb, cv2.COLOR_BGR2GRAY)
        return img_source_gray, img_search_rgb


class PicColorCanny(PicProcess):
    def __init__(self, threshold1=100, threshold2=200):
        super().__init__('背景去除')
        self.threshold1 = threshold1  # 用于边缘连接。如果边缘强度大于这个值，Canny 算法将其分类为边缘
        self.threshold2 = threshold2  # 当边缘强度小于此值时，如果边缘与强边缘相连，则该边缘也将被接受

    def process(self, img_source_rgb: np.ndarray, img_search_rgb: np.ndarray) -> (np.ndarray, np.ndarray):
        img_source_rgb = cv2.Canny(img_source_rgb, self.threshold1, self.threshold2)
        img_search_rgb = cv2.Canny(img_search_rgb, self.threshold1, self.threshold2)
        return img_source_rgb, img_search_rgb


class PicCrop(PicProcess):
    def __init__(self, src_crop_box: Box):
        super().__init__('图片裁剪')
        self.src_crop_box = src_crop_box

    def process(self, img_source_rgb: np.ndarray, img_search_rgb: np.ndarray) -> (np.ndarray, np.ndarray):
        height, width = img_source_rgb.shape[:2]
        x1, y1, x2, y2 = self.src_crop_box.tuple()
        x1 = max(0, x1)
        y1 = max(0, y1)
        x2 = min(width, x2)
        y2 = min(height, y2)

        img_source_crop = img_source_rgb[y1:y2, x1:x2]
        return img_source_crop, img_search_rgb



from typing import List

from auto_easy.dag.executor import Executor
from auto_easy.dag.layer import DAGLayerDef, LayerConf, DAGLayerSimple
from auto_easy.models import Ctx
from auto_easy.utils import is_actual_subclass, logger


class DAG(Executor):
    def __init__(self, name, retry_mode=None):
        self.layers: List[DAGLayerDef] = []
        self.retry_mode = retry_mode  # None-不重试, 1-回退重试
        super().__init__(name)

    # @abstractmethod
    def init(self):
        pass

    def hit(self, ctx: Ctx) -> bool:
        if len(self.layers) == 0:
            raise Exception("Dag empty, no layers added, name: {}".format(self.name))
        return self.layers[0].hit(ctx=ctx)

    def exec(self, ctx: Ctx) -> bool:
        if len(self.layers) == 0:
            raise Exception("Dag empty, no layers added, name: {}".format(self.name))

        logger.debug(f'开始执行DAG: {self.name}')
        idx = 0
        retry = 0
        while idx < len(self.layers):
            layer = self.layers[idx]
            logger.debug(layer.name)
            succ = False
            if layer.hit(ctx=ctx):
                if layer.run(ctx=ctx):
                    succ = True
            if succ or layer.conf.skip_err:
                idx += 1
                continue

            if self.retry_mode and self.retry_mode == 1 and retry < 1:
                if idx > 0:
                    logger.warning("Retrying layer, failed layer: {}, goback: {}".format(self.layers[idx].name,
                                                                                         self.layers[idx - 1].name))
                    idx = idx - 1
                    retry += 1
                    continue

            logger.debug(f'DAG({self.name}) 执行失败, layer: {layer.name}')
            return False

        logger.debug(f'结束执行DAG: {self.name}')
        return True

    def add_layer(self, layer, conf: LayerConf = None):
        if not is_actual_subclass(layer, DAGLayerDef):
            if is_actual_subclass(layer, Executor):
                layer = DAGLayerSimple(executor=layer, conf=conf)

        if not is_actual_subclass(layer, DAGLayerDef):
            raise Exception("Dag add layer failed")
        self.layers.append(layer)

    def add_layers(self, layers):
        for layer in layers:
            self.add_layer(layer)

    @staticmethod
    def simple_new(name, executors: List[Executor]):
        dag = DAG(name)
        dag.add_layers(executors)
        return dag


class SubDAG2Executor(Executor):
    def __init__(self, dag: DAG):
        super().__init__(dag.name)
        self.dag = dag

    def hit(self, ctx: Ctx) -> bool:
        return self.dag.hit(ctx=ctx)

    def exec(self, ctx: Ctx) -> bool:
        return self.dag.run(ctx=ctx)

    @staticmethod
    def cvt(dag: DAG):
        return SubDAG2Executor(dag=dag)


if __name__ == "__main__":
    pass



from unittest import TestCase

from auto_easy import DAGLayerSwitchOne, ExecutorDebug, Ctx


class TestDAGLayerSwitchOne(TestCase):
    def test_exec(self):
        layer = DAGLayerSwitchOne(loop_to=1.1)
        layer.add_branch(ExecutorDebug('1', hit_ret=False, hit_wait=0.2))
        layer.add_branch(ExecutorDebug('2', hit_ret=False, hit_wait=0.2))
        layer.add_branch(ExecutorDebug('3', hit_ret=True, hit_wait=0.1, exec_ret=True))
        ok = layer.run(Ctx())
        self.assertTrue(ok)



import os

TestPicDir = os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
    'test_pics')


def get_test_pic(pic_name):
    return os.path.join(TestPicDir, pic_name)


def gen_test_pic(pic_name):
    return os.path.join(TestPicDir, pic_name)



from auto_easy.base.find_pic.conf_base import *
from auto_easy.base.find_pic.find_pic import *
from auto_easy.base.find_pic.model import *
from auto_easy.base.find_pic.pic_factory import *
from auto_easy.base.find_pic.pic_handler import *



import os

import PIL
import cv2
import numpy as np
from PIL import Image


def img_2_pil(img):
    if isinstance(img, str):
        if not os.path.exists(img):
            raise FileNotFoundError(img)
        return Image.open(img)
    if isinstance(img, np.ndarray):
        return Image.fromarray(img)
    if isinstance(img, PIL.Image.Image):
        return img
    raise TypeError("img must be str or PIL.Image.Image")


def img_2_ndarray_rgb(img):
    if isinstance(img, str):
        if not os.path.exists(img):
            raise FileNotFoundError(img)
        image_bgr = cv2.imread(img)
        image_rgb = cv2.cvtColor(image_bgr, cv2.COLOR_BGR2RGB)
        return image_rgb

    if isinstance(img, PIL.Image.Image):
        return np.array(img)

    if isinstance(img, np.ndarray):
        return img
    raise TypeError('img must be PIL.Image.Image or np.ndarray, type: {}'.format(type(img)))


def img_2_ndarray_gbr(img):
    if isinstance(img, str):
        return cv2.imread(img)
    if isinstance(img, PIL.Image.Image):
        rgb_array = np.array(img)
        bgr_array = cv2.cvtColor(rgb_array, cv2.COLOR_RGB2BGR)
        return bgr_array
    if isinstance(img, np.ndarray):
        return img
    raise TypeError('img must be PIL.Image.Image or np.ndarray, type: {}'.format(type(img)))


def img_2_list(img):
    ndarray = img_2_ndarray_rgb(img)
    return ndarray.tolist()


def list_2_ndarray(list_img):
    return np.array(list_img)
    # ndarray = img_2_ndarray(list_img)

    # def rpc_api_predict(self,model_name,image_bytes) -> {} :
    #     self.set_cur_model(model_name)
    #     image = Image.open(io.BytesIO(image_bytes.data))
    #     print('[client_rpc_call] rpc api predict, image_size = {}'.format(image.size))
    #     results = self.cur_model.predict(image)
    #     json.dumps(cls_to_dict(results))
    #     return json.dumps(cls_to_dict(results))
    #
    # def rpc_call_predict(self, model_name, image) -> YoloObjs:
    #     if isinstance(image, str):
    #         image = Image.open(image)
    #     image_bytes = io.BytesIO()
    #     image.save(image_bytes, format='PNG')
    #     image_bytes = image_bytes.getvalue()
    #     bs = self.rpc_client.ocr_rpc_api_predict(model_name, image_bytes)
    #     d = json.loads(bs)



from auto_easy.base.find_color import find_color, find_best_color
from auto_easy.base.windows import Window


class WinFindColor(Window):
    def __init__(self, window_id=None, pic_dir=None):
        Window.__init__(self, window_id=window_id)

    def find_color(self, color, box=None) -> bool:
        img = self.capture_window()
        if box is not None:
            img = img.crop(box.tuple())
        exists, cover_rate = find_color(img, color)
        return exists

    def find_most_color(self, colors_offset, min_area_rate=0.0, box=None):
        screen = self.capture(box)
        return find_best_color(screen, colors_offset, min_area_rate)



import datetime
import os

from auto_easy.base.image.draw import draw_rectangles
from auto_easy.base.windows import Window
from auto_easy.models import Box
from auto_easy.utils import logger


class WinShow(Window):
    def __init__(self, window_id=None, pic_save_dir=None):
        Window.__init__(self, window_id=window_id)
        self.pic_save_dir = pic_save_dir

    def show(self, boxes: list[Box] = None):
        img = self.capture_window()
        draw_img = draw_rectangles(img, boxes, use_name=True)
        draw_img.show()

    def save(self, prefix, img=None, time_suffix=True, ext='bmp', debug_print=True):
        if self.pic_save_dir is None:
            raise Exception('save_dir must be set')
        prefix = prefix.replace("\\", "").replace("/", "")
        target_dir = os.path.join(self.pic_save_dir, self.get_text())
        if not os.path.exists(target_dir):
            os.mkdir(target_dir)
        now = datetime.datetime.now()
        if time_suffix:
            file_time_str = now.strftime('%Y%m%d_%H%M%S_%f')[:-3]
            file_name = f'{prefix}_{file_time_str}.{ext}'
        else:
            file_name = f'{prefix}.{ext}'
        path = os.path.join(target_dir, file_name)
        if img is None:
            img = self.capture(latest_lag=3)
        img.save(path)
        if debug_print:
            logger.debug('save screenshot to: {}'.format(path))


if __name__ == '__main__':
    win = WinShow('Phone-VB')
    win.show([Box(10, 10, 100, 100)])
    win.pic_save_dir = '.'
    win.save('test', debug_print=True)

    # win.pic_dir = TestPicDir
    # start = time.time()
    # mdet = win.loop_find_pics('core/test_1', to=10, sleep=1)
    # mdet = win.find_pics('core/test_1')
    # print(mdet)
    # print(time.time() - start)

    # exists = win.exists_pics('core/test_1')
    # print(exists)



# from datetime import time

import time

from auto_easy.global_log import logger


def timeit_decorator(func):
    def wrapper(*args, **kwargs):
        start_time = time.time()
        # 获取函数名称
        func_name = func.__name__
        result = func(*args, **kwargs)
        end_time = time.time()
        elapsed_time = end_time - start_time
        logger.info(f"函数 {func_name} 执行完毕，耗时: {elapsed_time:.3f} 秒")
        return result

    return wrapper



import os

from auto_easy.base.find_pic.model import PicV2
from auto_easy.constant import TestPicDir
from auto_easy.utils import concurrent_exec_one_func
from auto_easy.utils.file_util import get_files, is_image


class PicFactory:
    def __init__(self, pic_dir=None):
        self.pic_list: list[PicV2] = []
        self._pic_dir = None
        if pic_dir:
            self.pic_dir = pic_dir

    @property
    def pic_dir(self):
        return self._pic_dir

    @pic_dir.setter
    def pic_dir(self, pic_dir):
        self._pic_dir = pic_dir
        self.load_all_pic()

    def _load_one_pic(self, path) -> PicV2:
        relative_path = os.path.relpath(path, self.pic_dir)
        relative_path = relative_path.replace("\\", "/")
        filename_without_ext = os.path.splitext(relative_path)[0]
        sp = filename_without_ext.split('$$$')
        if len(sp) < 1:
            raise Exception("图片命名格式错误: {}".format(path))
        pic_name = sp[0]
        return PicV2(name=pic_name, path=path)

    def load_all_pic(self):
        files = get_files(self.pic_dir)
        pic_file_list = [path for path in files if is_image(path)]
        pic_list = concurrent_exec_one_func(self._load_one_pic, pic_file_list)
        self.pic_list = pic_list

    def get_pic(self, name) -> PicV2:
        name = name.replace("\\", "/")
        for pic in self.pic_list:
            if pic.name == name:
                return pic
        return None

    def get_pics(self, name) -> list[PicV2]:
        name = name.replace("\\", "/")
        ans = []
        for pic in self.pic_list:
            if pic.name == name:
                ans.append(pic)
        return ans

    def get_pics_by_prefix(self, prefix) -> list[PicV2]:
        prefix = prefix.replace("\\", "/")
        ans = []
        for pic in self.pic_list:
            if pic.name.startswith(prefix):
                ans.append(pic)
        return ans

    def get_all_pics(self, prefix_excludes=None) -> list[PicV2]:
        prefix_excludes = [] if prefix_excludes is None else prefix_excludes
        ans = []
        for pic in self.pic_list:
            filter = False
            for pic_exclude in prefix_excludes:
                if pic.name.startswith(pic_exclude):
                    filter = True
                    break
            if not filter:
                ans.append(pic)

        return ans


if __name__ == '__main__':
    pic_factory = PicFactory()
    pic_factory.pic_dir = TestPicDir
    for pic in pic_factory.pic_list:
        print(pic)

    print(pic_factory.get_pic('find_pic\search_pic_1'))



from auto_easy.utils.cache_util import *
from auto_easy.utils.cls_util import *
from auto_easy.utils.common import *
from auto_easy.utils.debug import *
from auto_easy.utils.file_util import *
from auto_easy.utils.list_util import *
from auto_easy.utils.multi_thread import *
from auto_easy.utils.queue_util import *
from auto_easy.utils.time_util import *



import builtins


def is_class_name(obj, cls):
    mro_list = obj.__class__.__mro__
    mro_name_list = [c.__name__ for c in mro_list]
    return cls.__name__ in mro_name_list


def is_builtins(obj):
    return isinstance(obj, builtins.object)


def cls_to_dict(obj):
    if not hasattr(obj, '__dict__'):
        return obj
    d = vars(obj)
    for k, v in d.items():
        if isinstance(v, list) and len(v) > 0 and is_builtins(v[0]):
            for idx, list_val in enumerate(v):
                v[idx] = cls_to_dict(list_val)
            continue
        if is_builtins(v):
            d[k] = cls_to_dict(v)
            continue
    return d


def set_obj_by_dict(obj, d):
    for k, v in obj.__dict__.items():
        if k in d:
            setattr(obj, k, d[k])


# 判断是对象还是类
def is_cls(v):
    return isinstance(v, type)


def is_actual_subclass(subclass, superclass):
    subclass = subclass
    if not is_cls(subclass):
        subclass = subclass.__class__
    # 判断 subclass 是否是 superclass 的实际子类
    return subclass == superclass or (issubclass(subclass, superclass) and superclass in subclass.__mro__)



import cv2

from auto_easy.base.image.cvt import img_2_ndarray_rgb, img_2_ndarray_gbr
from auto_easy.constant import get_test_pic


def parse_color_range_bgr(color_deviation):
    """
    解析大漠偏色范围 (格式: 主色-偏差)
    Args:
        color_deviation (str): 偏色字符串，例如 "C28D3F-3D333E"

    Returns:
        tuple: 下限颜色范围 (BGR), 上限颜色范围 (BGR)
    """
    # 分割主色和偏差
    main_color_hex, deviation_hex = color_deviation.split('-')

    # 将主色和偏差转换为 BGR
    main_color = tuple(int(main_color_hex[i:i + 2], 16) for i in (4, 2, 0))  # 转为BGR顺序
    deviation = tuple(int(deviation_hex[i:i + 2], 16) for i in (4, 2, 0))  # 偏差同样转为BGR顺序

    # 计算下限和上限
    lower_bound = np.clip([m - d for m, d in zip(main_color, deviation)], 0, 255).astype(np.uint8)
    upper_bound = np.clip([m + d for m, d in zip(main_color, deviation)], 0, 255).astype(np.uint8)

    return lower_bound, upper_bound


def img_binarize(img, color_deviation, inverted=False):
    img = img_2_ndarray_rgb(img)

    lower_bound, upper_bound = parse_color_range_bgr(color_deviation)

    # 转为二值化
    binary_img = cv2.inRange(img, lower_bound, upper_bound)
    if inverted:
        inverted = cv2.bitwise_not(binary_img)
        return inverted
    return binary_img


def image_resize(img, scale_factor):
    new_size = (int(img.shape[1] * scale_factor), int(img.shape[0] * scale_factor))
    print(img.shape[1], img.shape[0])
    resized_img = cv2.resize(img, new_size, interpolation=cv2.INTER_CUBIC)
    print(resized_img.shape[1], resized_img.shape[0])
    return resized_img


def image_color_keep(img, color_deviation):
    img = img_2_ndarray_gbr(img)
    lower_bound, upper_bound = parse_color_range_bgr(color_deviation)

    # 创建掩模，保留在颜色范围内的所有像素
    result_image = np.zeros_like(img)
    color_mask = cv2.inRange(img, lower_bound, upper_bound)
    # 只保留掩模中为白色的部分
    result_image[color_mask > 0] = img[color_mask > 0]
    return result_image


import numpy as np
from skimage.metrics import structural_similarity as ssim


def compare_images(image1, image2):
    # 将 PIL Image 转换为 NumPy 数组
    img1 = img_2_ndarray_rgb(image1)
    img2 = img_2_ndarray_rgb(image2)

    # 检查两张图片的尺寸是否相同
    if img1.shape != img2.shape:
        raise ValueError("Images must have the same dimensions for SSIM.")

    # 计算 SSIM
    similarity_index, _ = ssim(img1, img2, win_size=3, full=True, multichannel=True)
    return similarity_index


def contain_color(img, color_range):
    """
    在图片中寻找符合条件的颜色范围，并计算面积占比
    :param image: 输入图片路径
    :param color_range: 偏差参数，格式如 "F7E97D-08123D"
    :return: 占比百分比
    """
    # 解析颜色区间
    lower_bound, upper_bound = parse_color_range_bgr(color_range)
    image = img_2_ndarray_gbr(img)

    # 根据颜色范围生成掩码
    mask = cv2.inRange(image, lower_bound, upper_bound)

    # 计算符合条件的像素数量
    matching_pixels = np.count_nonzero(mask)

    # 计算图片总像素数量
    total_pixels = image.shape[0] * image.shape[1]

    # 是否存在符合条件的颜色
    exists = matching_pixels > 0

    # 计算面积占比（百分比输出）
    area_percentage = (matching_pixels / total_pixels) * 100

    return round(area_percentage, 2)


if __name__ == '__main__':
    # img = image_color_keep()
    print(contain_color(get_test_pic('debug/cs.bmp'), 'BF7600-080700'))
    print(contain_color(get_test_pic('debug/wfcs.bmp'), 'BF7600-080700'))
    # path = r'E:\repo\dnf_tool\biz\det_pic\pics\选择角色\疲劳图标$$$box=86,324,96,338,宽高(10,14).bmp'
    # if not os.path.exists(path):
    #     print(f"File does not exist: {path}")
    # else:
    #     print(f"File found: {path}")
    # img = cv2.imread(path)
    # print(img)

    # img = image_color_keep(path,color_deviation='C4867F-3B777E')
    # cv2.imshow('img', img)
    # cv2.waitKey(0)
    # cv2.destroyAllWindows()
    # cv2.imwrite(gen_test_pic('lp'), img)
    # similarity_score = compare_images(get_test_pic('xlc.bmp'), get_test_pic('xlc.bmp'))
    # print("Images similarity score (SSIM):", similarity_score)
    # pass
    # sys.exit()
    # 读取输入图像
    # ori_path = 'test_pics/upscaled_image.jpg'
    # resize_path = 'test_pics/resized_img.bmp'
    # binary_path = 'test_picbinary_img.bmp'
    # img = cv2.imread(ori_path)
    #
    # ans = ai_ocr.predict(ori_path, scale_factor=1, allow_rpc=True)
    # print(ans)
    #
    # # 颜色偏差（例如来自大漠偏色计算器）
    # print('放大处理')
    # color_deviation = 'A27842-483B42'
    # color_deviation = '502A0D-482A0E' # 背景色
    # resized_img = image_resize(img, 10)
    # cv2.imwrite(resize_path, resized_img)
    # ans = ai_ocr.predict(resize_path, scale_factor=1, allow_rpc=True)
    # print(ans)
    #
    # # 调用函数处理
    # print('二值化处理')
    # binary_img = img_binarize(resized_img, color_deviation)
    # cv2.imwrite(binary_path, binary_img, )
    # ans = ai_ocr.predict(binary_path, scale_factor=1, allow_rpc=True)
    # print(ans)
    #
    # 保存结果
    # print("二值化处理完成，图像已保存为 binary_image.png")
    #
    # print("二值化处理完成，图像已保存为 resized_img2.png")



import time
from typing import List

from auto_easy.models import Box, sort_and_group_points


class Pic:
    def __init__(self, path):
        self.path = path  # 完整文件路径
        self.name = ""  # 文件名
        self.name_final = ""  # 文件名中最后的名称
        self.cut_box = None  # 游戏截图区域
        self.click_area_rate = None  # 点击区域,百分比,如(0.5,0.5,1,1)
        self.delta_color = ''
        self.width = 0  # 宽
        self.height = 0  #
        self.params = {}
        self.pil_img = None
        self.ndarray_obj = None
        self.sim = -1

    def is_same(self, name) -> bool:
        return self.name == name

    def gen_click_box(self, match_box: Box):
        if not self.click_area_rate:
            return match_box
        new_box = match_box.crop_by_rate(self.click_area_rate[0], self.click_area_rate[1],
                                         self.click_area_rate[2], self.click_area_rate[3])
        return new_box

    def __str__(self):
        return '{}{}'.format(self.name, self.cut_box)


class PicDet:
    def __init__(self, pic: Pic, boxes: List[Box]):
        self.pic = pic
        self.is_detected = False
        self.match_boxes = []
        self.ts = time.time()
        self.add_boxes(boxes)

    @property
    def match_box_first(self) -> Box:
        return self.match_boxes[0]

    def replace_boxes(self, boxes: List[Box]):
        self.match_boxes = boxes
        self.is_detected = len(self.match_boxes) > 0

    # TODO： 待廢棄
    def add_boxes(self, boxes: List[Box]):
        self.match_boxes += boxes
        self.is_detected = len(self.match_boxes) > 0

    def get_boxes(self, dir=1, dis_err=20):
        # TODO: 支持按不同顺序排序, 现在默认是从左往右，从上往下
        points = [box.tl_point for box in self.match_boxes]
        sorted_points = sort_and_group_points(points, dis_err=dis_err)
        # points的顺序映射到box的顺序
        boxes = []
        for point in sorted_points:
            for box in self.match_boxes:
                if id(box.tl_point) == id(point):
                    boxes.append(box)

        return boxes

    def get_click_box(self):
        if not self.is_detected:
            return None
        box = self.match_box_first
        box = self.pic.gen_click_box(box)  # 图片有参数可以指定点击区域
        return box

    def __str__(self):
        return '{}：{}'.format(self.pic.name, '|'.join([str(box) for box in self.match_boxes]))


class MPicDet:
    def __init__(self, pics_det: List[PicDet] = []):
        self.results = pics_det
        # self.is_detected = len(pics_det) > 0
        self.create_ts = time.time()

    @property
    def is_detected(self):
        for det in self.results:
            if det.is_detected:
                return True
        return False

    def get_output_exists_names(self) -> list[str]:
        ans = []
        for det in self.results:
            if det.is_detected:
                ans.append(det.pic.name)
        return ans

    def get_output_first(self) -> PicDet:
        for det in self.results:
            if det.is_detected:
                return det

    def exists_inputs(self, names: list[str]) -> bool:
        names = names if isinstance(names, list) else [names]
        for name in names:
            det = self.get(name)
            if det is None:
                return False
        return True

    def get(self, name) -> PicDet:
        for v in self.results:
            if v.pic.is_same(name):
                return v
        return None

    def get_box(self, name) -> Box:
        return self.get(name).match_box_first

    def exists_output(self, names: list[str]) -> bool:
        if not self.is_detected:
            return False
        names = names if isinstance(names, list) else [names]
        for name in names:
            det = self.get(name)
            if det is None or det.is_detected == False:
                return False
        return True

    def not_exists_output(self, names: list) -> bool:
        if not self.is_detected:
            return True
        names = names if isinstance(names, list) else [names]
        for name in names:
            det = self.get(name)
            if det is not None and det.is_detected:
                return False
        return True

    def merge_other(self, other):
        assert isinstance(other, MPicDet)
        for other_det in other.results:
            old_det = self.get(other_det.pic.name)

            if old_det is None:
                self.results.append(other_det)
            else:
                old_det.replace_boxes(other_det.match_boxes)

    def merge_others(self, others):
        assert isinstance(others, list)
        for other in others:
            self.merge_other(other)

    def add_pics_det(self, pics_det: List[PicDet]):
        if isinstance(pics_det, MPicDet):
            pics_det = pics_det.results

        for new_det in pics_det:
            old_det = self.get(new_det.pic.name)
            if old_det is None:
                self.results.append(new_det)
            else:
                old_det.add_boxes(new_det.match_boxes)

    def print_results(self):
        tmp = [str(pic) for pic in self.results]
        return '|'.join(tmp)

    def print_exists_results(self):
        tmp = []
        for det in self.results:
            if det.is_detected:
                tmp.append(str(det))
        return '|'.join(tmp)

    def get_all_boxes(self):
        boxes = []
        for v in self.results:
            boxes = boxes + v.match_boxes
        return boxes



import copy
import hashlib
import os
import re

import PIL
import cv2
import numpy as np
from PIL import Image

from auto_easy.base.find_pic.conf_base import ConfBase
from auto_easy.base.image.draw import draw_rectangles
from auto_easy.constant import get_test_pic
from auto_easy.models import Box, sort_boxes_by_group, group_box

PicDetConf_box = None
PicDetConf_expand_scale = 2  # 匹配区域box缩放比例, 1表示不缩放, 2表示放大两倍
PicDetConf_multi_match = False  # 是否允许一张图片匹配多个
PicDetConf_multi_dedup = True  # 匹配多个的时候进行去重, 保留sim最高
PicDetConf_find_one = False  # 只返回一个
PicDetConf_color = ""  # 该参数如果设置了非空，会覆盖图片名称中的偏色
PicDetConf_contain_color = ""  # 该参数如果设置了非空，会覆盖图片名称中的偏色
PicDetConf_sim = 0.8
PicDetConf_range_scale = (0.98, 1.02, 0.01)  # 自动大小匹配, 左闭右闭区间, 格式 (start, end, step) , 例如(0.95, 1.1, 0.1)
PicDetConf_cur_scale = None
PicDetConf_rgb = True  # 是否在RPB模式下匹配
PicDetConf_bg_remove = False
PicDetConf_method = cv2.TM_CCOEFF_NORMED
PicDetConf_debug = False


class PicDetConf(ConfBase):
    def __init__(self):
        super().__init__()
        self._box: Box = PicDetConf_box  # 匹配区域, None/Box(0,0,0,0) 默认全屏
        self._expand_scale = PicDetConf_expand_scale  # 匹配区域box缩放比例, 1表示不缩放, 2表示放大两倍
        self._multi_match = PicDetConf_multi_match  # 是否允许一张图片匹配多个
        self._multi_dedup = PicDetConf_multi_dedup  # 匹配多个的时候进行去重, 保留sim最高
        self._find_one = PicDetConf_find_one  # 只返回一个
        self._color = PicDetConf_color  # 该参数如果设置了非空，会覆盖图片名称中的偏色
        self._contain_color = PicDetConf_contain_color  # 必须包含的颜色
        self._sim = PicDetConf_sim
        self._range_scale = PicDetConf_range_scale  # 自动大小匹配, 左闭右闭区间, 格式 (start, end, step) , 例如(0.95, 1.1, 0.1)
        self._cur_scale = PicDetConf_cur_scale
        self._rgb = PicDetConf_rgb  # 是否在RPB模式下匹配
        self._bg_remove = PicDetConf_bg_remove
        self._method = PicDetConf_method
        self._debug = PicDetConf_debug
        self._params = {}

    @property
    def cur_scale(self):
        return self._cur_scale

    @cur_scale.setter
    def cur_scale(self, value):
        self._cur_scale = float(value)

    @property
    def box(self):
        return self._box

    @property
    def scaled_box(self):
        if self.box is None:
            return None
        origin_box = self.box
        if self.expand_scale is not None and self.expand_scale != 1:
            scaled_box = self._box.copy_by_scale(self.expand_scale)
            return scaled_box
        return origin_box

    @box.setter
    def box(self, v):
        if v is None:
            self._box = None
            return
        if isinstance(v, Box):
            self._box = v
            return
        if isinstance(v, str):
            s = v
            if len(s) == 0:
                self._box = None
                return
            ls = s.split(',')
            if len(ls) >= 4:
                self._box = Box(ls[0], ls[1], ls[2], ls[3])
                return
        raise Exception('invalid box, {}'.format(v))

    @property
    def expand_scale(self):
        return self._expand_scale

    @expand_scale.setter
    def expand_scale(self, v):
        self._expand_scale = int(v)

    @property
    def multi_match(self):
        return self._multi_match

    @multi_match.setter
    def multi_match(self, v):
        self._multi_match = self._cvt_bool(v)

    @property
    def find_one(self):
        return self._find_one

    @find_one.setter
    def find_one(self, v):
        self._find_one = self._cvt_bool(v)
        if self._find_one is None:
            raise Exception('invalid find one, {}'.format(v))

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, v):
        self._color = str(v)

    @property
    def contain_color(self):
        return self._contain_color

    @contain_color.setter
    def contain_color(self, v):
        self._contain_color = str(v)

    @property
    def sim(self):
        return self._sim

    @sim.setter
    def sim(self, v):
        self._sim = float(v)
        if not (0 <= self._sim <= 1):
            raise Exception('invalid sim, {}'.format(v))

    @property
    def range_scale(self):
        return self._range_scale

    @range_scale.setter
    def range_scale(self, v):
        self._range_scale = None
        if v is None:
            return
        if isinstance(v, tuple) and len(v) == 3:
            self._range_scale = (float(v[0]), float(v[1]), float(v[2]))

        if isinstance(v, str):
            if len(v) == 0:
                self._range_scale = (1, 1, 0)
            ls = v.split(',')
            if len(ls) == 3:
                self._range_scale = (float(ls[0]), float(ls[1]), float(ls[2]))

        if self._range_scale is None:
            raise Exception('invalid range_scale, {}'.format(v))

        if self._range_scale[0] > self._range_scale[1] or self._range_scale[1] - self._range_scale[0] < \
                self._range_scale[2]:
            raise Exception('invalid range_scale, {}'.format(v))
        # print(self._range_scale)

    @property
    def rgb(self):
        return self._rgb

    @rgb.setter
    def rgb(self, v):
        self._rgb = self._cvt_bool(v)
        if self._rgb is None:
            raise Exception('invalid rgb, {}'.format(v))

    @property
    def bg_remove(self):
        return self._bg_remove

    @bg_remove.setter
    def bg_remove(self, v):
        self._bg_remove = self._cvt_bool(v)

    @property
    def method(self):
        return self._method

    @method.setter
    def method(self, v):
        _method = int(v)
        template_matching_methods = [
            cv2.TM_SQDIFF,
            cv2.TM_SQDIFF_NORMED,
            cv2.TM_CCORR,
            cv2.TM_CCORR_NORMED,
            cv2.TM_CCOEFF,
            cv2.TM_CCOEFF_NORMED
        ]
        if _method not in template_matching_methods:
            raise Exception('invalid method, {}'.format(v))
        self._method = _method

    @property
    def debug(self):
        return self._debug

    @debug.setter
    def debug(self, v):
        self._debug = bool(v)

    @property
    def params(self):
        return self._params

    @params.setter
    def params(self, v):
        self._params = v

    def _cvt_bool(self, v):
        if isinstance(v, bool):
            return v
        if v in ['True', 'true', '1']:
            return True
        if v in ['False', 'false', '0']:
            return False
        raise Exception('invalid cvt_bool, {}'.format(v))

    def update_with_kv_str(self, kv_str: str):
        # kv_str例子: box=0,0,1,1&sim=0.8
        items = kv_str.split('&')
        params = {}
        for item in items:
            key, value = item.split('=')
            params[key] = value
            if not hasattr(PicDetConf, key):
                continue
            if isinstance(getattr(PicDetConf, key), property):
                setattr(self, key, value)
        setattr(self, 'params', params)

    def deepcopy(self):
        obj = copy.deepcopy(self)
        return obj


class PicV2:
    def __init__(self, name='', path='', pil_img=None, cv2_img=None, det_conf: PicDetConf = None):

        self.pil_img_rgb = None
        self.cv2_img_bgr = None
        # 产出图片数据
        if path != '':
            self._is_image(path)
            self.pil_img_rgb = Image.open(path)
            # cv2 imread不支持中文路径, 通过pil转化
            self.cv2_img_bgr = cv2.cvtColor(np.array(self.pil_img_rgb), cv2.COLOR_RGB2BGR)
        elif pil_img is not None:
            self.pil_img_rgb = pil_img
            self.cv2_img_bgr = cv2.cvtColor(np.array(pil_img), cv2.COLOR_RGB2BGR)
        elif cv2_img is not None:
            self.cv2_img_bgr = cv2_img
            self.pil_img_rgb = Image.fromarray(cv2.cvtColor(cv2_img, cv2.COLOR_BGR2RGB))
        else:
            raise Exception('path or pil_img or cv2_img are required')

        self.path = path
        self.name = ''  # 优先级: name > path > image.sha256
        if name != '':
            self.name = name
        elif path != '':
            self.name = path
        else:
            hasher = hashlib.sha256()
            hasher.update(self.cv2_img_bgr)
            self.name = hasher.hexdigest()

        base_conf = PicDetConf()  # 默认配置
        file_conf = PicDetConf()  # 文件名中的配置
        user_conf = det_conf  # 用户传入的配置
        if path != '':
            file_name_without_extension = os.path.splitext(os.path.basename(path))[0]
            ls = file_name_without_extension.split('$$$')
            if len(ls) < 1:
                raise Exception('invalid path, {}'.format(path))
            if len(ls) > 1:
                file_conf.update_with_kv_str(ls[1])
        self.det_conf = ConfBase.new_conf_by_pry([base_conf, file_conf, user_conf])  # 按低优到高优进行配置覆盖

    def _is_image(self, path) -> bool:
        if not os.path.exists(path):
            raise FileNotFoundError(path)
        try:
            img = Image.open(path)  # 尝试打开图像文件
            img.verify()  # 验证文件是否损坏
            return True
        except (IOError, SyntaxError):  # PIL无法识别的文件类型会抛出IOError
            raise Exception('path exists, but file is not image, {}'.format(path))

    def deepcopy(self):
        obj = copy.deepcopy(self)
        return obj

    @staticmethod
    def new_auto(v, name=''):
        if isinstance(v, PicV2):
            return v

        # 传入文件路径
        if isinstance(v, str):
            return PicV2(name=name, path=v)

        # 传入PIL.Image
        if isinstance(v, PIL.Image.Image):
            return PicV2(name=name, pil_img=v)

        # 传入cv2的ndarray
        if isinstance(v, np.ndarray):
            return PicV2(name=name, cv2_img=v)
        raise Exception(
            'Piv new_auto only support str/PIL.Image.Image/np.ndarray, input: {}, type: {}'.format(v, type(v)))

    def __str__(self):
        return self.path

    def show(self, boxes=None):
        draw_img = self.pil_img_rgb
        if boxes:
            draw_img = draw_rectangles(self.pil_img_rgb, boxes=boxes)
        draw_img.show()


class DetBox(Box):
    def __init__(self, sim, x1, y1, x2, y2, name=''):
        super().__init__(x1, y1, x2, y2, name)
        self.sim = float(sim)

    @staticmethod
    def new_det_boxes_by_str(s):
        # 使用正则表达式提取<x1,y1,x2,y2>
        pattern = r'(\d+\.\d+)<(\d+),(\d+),(\d+),(\d+)>'
        matches = re.findall(pattern, s)
        # 打印提取的坐标
        boxes = []
        for match in matches:
            score, x1, y1, x2, y2 = match
            boxes.append(DetBox(score, x1, y1, x2, y2, str(score)))
        return boxes

    def __str__(self):
        return f'{self.sim}<{self.x1},{self.y1},{self.x2},{self.y2}>'


class PicDetV2:
    def __init__(self, pic: PicV2, det_boxes: list[DetBox] = None, scale=1):
        self.pic = pic
        self.scale = scale
        self.boxes: list[DetBox] = []
        self.add_boxes(det_boxes)
        # self.boxes = det_boxes if det_boxes is not None else []

    def replace_boxes(self, det_boxes: list[DetBox]):
        self.boxes = []
        self.add_boxes(det_boxes)

    def add_box(self, det_box: DetBox):
        self.boxes.append(det_box)

    def add_boxes(self, det_boxes: list[DetBox]):
        self.boxes.extend(det_boxes)
        self.boxes = sorted(self.boxes, key=lambda box: box.sim, reverse=True)

    @property
    def is_detected(self):
        return len(self.boxes) > 0

    @property
    def box(self):
        if self.boxes:
            return self.boxes[0]
        return None

    def get_boxes(self, sort_mode=0, dis_err=10):
        # 默认按置信度排序
        if sort_mode == 0:
            return self.boxes
        if sort_mode == 1:
            return sort_boxes_by_group(self.boxes, dis_err=dis_err)
        raise Exception('sort_mode must be 0 or 1')

    def get_boxes_group(self, dis_err):
        boxes = self.boxes
        return sort_boxes_by_group(boxes, dis_err=dis_err)

    def group_boxes(self, dis_err) -> list[list[Box]]:
        boxes = self.boxes
        return group_box(boxes, dis_err=dis_err)

    def __str__(self):
        str_list = []
        for box in self.boxes:
            str_list.append(f'{box}')

        s = '|'.join(str_list)
        if len(self.pic.name) > 0:
            s = f'{self.pic.name}:{s}'
        return s

    def merge_other(self, other, dedup=False):
        if len(other.boxes) == 0:
            return
        self.add_boxes(other.boxes)
        if dedup:
            self.boxes = nms_boxes(self.boxes)

    def merge_others(self, others, dedup=False):
        for other in others:
            self.merge_other(other, dedup)


def nms_boxes(det_boxes):
    # 准备进行非极大值抑制
    boxes = np.array([(box.x1, box.y1, box.width, box.height) for box in det_boxes])
    scores = np.array([box.sim for box in det_boxes])

    # 执行非极大值抑制
    indices = cv2.dnn.NMSBoxes(boxes.tolist(), scores.tolist(), score_threshold=0.5, nms_threshold=0.1)

    # 存储经过 NMS 处理后的匹配结果
    filter_boxes = [det_boxes[i] for i in indices]
    return filter_boxes


class MPicDetV2:
    def __init__(self, pic_det_list: list[PicDetV2] = None):
        self.pic_det_list = pic_det_list if pic_det_list is not None else []

    @property
    def first_det(self):
        for det in self.pic_det_list:
            if det.is_detected:
                return det
        return None

    @property
    def is_detected(self):
        for pic_det in self.pic_det_list:
            if pic_det.is_detected:
                return True
        return False

    @property
    def all_detected(self):
        for det in self.pic_det_list:
            if not det.is_detected:
                return False
        return True

    @property
    def boxes(self) -> list[DetBox]:
        _boxes = []
        for pic_det in self.pic_det_list:
            _boxes += pic_det.boxes
        return _boxes

    def get_boxes(self, pic=None, sort_mode=0, dis_err=10):
        boxes = self.boxes
        # todo: 支持指定具体的图片

        # 默认按置信度排序
        if sort_mode == 0:
            return sorted(self.boxes, key=lambda box: box.sim, reverse=True)

        if sort_mode == 1:
            return sort_boxes_by_group(boxes, dis_err=dis_err)

    @property
    def box(self) -> DetBox:
        _boxes = self.boxes
        if len(_boxes) == 0:
            return None
        return _boxes[0]

    def get_output_exists_names(self):
        return [det.pic.name for det in self.pic_det_list if det.is_detected]

    def merge_other(self, other, dedup=False):
        for det in other.pic_det_list:
            if dedup and self.get(det.pic.name):
                self.get(det.pic.name).merge_other(det, dedup)
            else:
                self.pic_det_list.append(det)

    def merge_others(self, others):
        for other in others:
            self.merge_other(other.pic_det_list)

    def get(self, name):
        for det in self.pic_det_list:
            if det.pic.name == name:
                return det
        return None

    def check(self, includes=None, excludes=None):
        includes = includes if includes is not None else []
        excludes = excludes if excludes is not None else []
        if not isinstance(includes, list):
            includes = [includes]
        for name in includes:
            det = self.get(name)
            if det is None or not det.is_detected:
                return False

        for name in excludes:
            det = self.get(name)
            if det is not None and det.is_detected:
                return False

        return True

    def merge_same_name(self, dedup=True):
        new_det_dict = {}
        for det in self.pic_det_list:
            if det.pic.name in new_det_dict.keys():
                new_det_dict[det.pic.name].merge_other(det, dedup)
            else:
                new_det_dict[det.pic.name] = det
        self.pic_det_list = list(new_det_dict.values())

    def filter_overlap(self):
        boxes = self.boxes
        boxes = nms_boxes(boxes)
        boxes_id = [id(box) for box in boxes]
        for det in self.pic_det_list:
            det_boxes = det.boxes
            filter_boxes = [box for box in det_boxes if id(box) in boxes_id]
            det.replace_boxes(filter_boxes)

    def print_exists_results(self):
        str_list = []
        for det in self.pic_det_list:
            if det.is_detected:
                str_list.append(det.__str__())
        return ';'.join(str_list)

    def __str__(self):
        str_list = []
        for det in self.pic_det_list:
            str_list.append(det.__str__())
        return ';'.join(str_list)


if __name__ == '__main__':
    # 从图片文件初始化
    path1 = get_test_pic('find_pic/search_pic_1$$$box=778,381,812,416.bmp')
    pic = PicV2.new_auto(path1, name='case1')
    print(pic.det_conf.__dict__)
    print(pic.name)

    # path2 = get_test_pic('find_pic/search_pic_2.bmp')
    # pic2 = PicV2.new_auto(path2, name='case2')
    # print(pic2.name)
    #
    # # 从cv2对象和PIL对象初始化
    # path3 = get_test_pic('find_pic/search_pic_1$$$box=778,381,812,416.bmp')
    # pil_img = Image.open(path3)
    # cv2_img = cv2.imread(path3)
    #
    # pic_from_cv2 = PicV2.new_auto(cv2_img)
    # pic_from_pil = PicV2.new_auto(pil_img)
    # print(pic_from_cv2.name)
    # print(pic_from_pil.name)
    # # pic3.show()
    # # pic4.show()
    #
    # if ImageChops.difference(pic_from_cv2.pil_img_rgb, pic_from_pil.pil_img_rgb).getbbox():
    #     print('pic3.pil_img_rgb != pic4.pil_img_rgb')
    #
    # if not np.array_equal(pic_from_cv2.cv2_img_bgr, pic_from_pil.cv2_img_bgr):
    #     print('pic3.pil_img_rgb != pic4.pil_img_rgb')



import time

from auto_easy.base.ai import AIModelBase
from auto_easy.base.ai import OCR
from auto_easy.base.ai.superres.ai_supper_res import SuperRes
from auto_easy.base.windows import Window
from auto_easy.constant import TestPicDir
from auto_easy.core.win_ai import WinYolo
from auto_easy.core.win_find_color import WinFindColor
from auto_easy.core.win_find_pic import WinFindPic
from auto_easy.core.win_monitor import WinTimer
from auto_easy.core.win_ocr import WinOCR
from auto_easy.core.win_show import WinShow
from auto_easy.models import Box


class CoreConf:
    def __init__(self, *args, **kwargs):
        self.pic_dir = ''
        self.window_id = ''  # 窗口名/窗口名前缀/窗口句柄
        self.pic_save_dir = None
        self.models: list[AIModelBase] = []
        self.rpc_server_port = 8765
        self.item_model_dir = ''  # 存储各种业务定制的Yolo Item类, 会自动进行转化


class AutoCore(
    WinYolo,
    WinOCR,
    WinFindPic,
    WinFindColor,
    WinShow,
    WinTimer):
    def __init__(self, conf: CoreConf):
        Window.__init__(self, window_id=conf.window_id)
        WinYolo.__init__(self, window_id=conf.window_id, models=conf.models, rpc_server_port=conf.rpc_server_port,
                         item_model_dir=conf.item_model_dir)
        WinOCR.__init__(self, window_id=conf.window_id, models=conf.models, rpc_server_port=conf.rpc_server_port)
        WinFindColor.__init__(self, window_id=conf.window_id)
        WinFindPic.__init__(self, window_id=conf.window_id, pic_dir=conf.pic_dir)
        WinShow.__init__(self, window_id=conf.window_id, pic_save_dir=conf.pic_save_dir)
        WinTimer.__init__(self, window_id=conf.window_id)


def interval_save(auto_core: AutoCore):
    auto_core.save('interval_save', debug_print=True)


_global_auto_core = None


def get_auto_core() -> AutoCore:
    global _global_auto_core
    if _global_auto_core is None:
        raise Exception('global_auto_core is not set')
    return _global_auto_core


def set_auto_core(auto_core: AutoCore):
    global _global_auto_core
    _global_auto_core = auto_core


if __name__ == '__main__':
    conf = CoreConf()
    conf.window_id = 'Phone-9a'
    conf.pic_dir = TestPicDir
    conf.pic_save_dir = TestPicDir + '/debug'

    supper_res = SuperRes()
    ocr = OCR()
    conf.models = [ocr, supper_res]
    # win = WinOCR('Phone-9a', models=[ocr, supper_res])
    # ans = win.ocr(Box(264, 53, 398, 81), scale_factor=2)

    auto_core = AutoCore(conf)
    print(auto_core.ocr(Box(264, 53, 398, 81), scale_factor=2))
    # auto_easy.add_timer_minitor(2, interval_save)

    time.sleep(10)
    # ans = auto_easy.find_pics('core/test_1')
    # print(ans)
    # auto_easy.window_id()
    # auto_easy.ocr()
    # auto_easy.find_pic()
    # auto_easy.a = 2
    # auto_easy.print_a1()
    # auto_easy.print_a2()



import cv2
import numpy as np

from auto_easy.base.ai.model_mgr_v2 import AIModelBase
from auto_easy.base.ai.superres.ai_supper_res import SuperRes
from auto_easy.base.image.cvt import img_2_ndarray_rgb, img_2_list, list_2_ndarray
from auto_easy.constant import get_test_pic


class OCR(AIModelBase):
    def __init__(self, *args, **kwargs):
        self.arg = args
        self.kwargs = kwargs
        super().__init__(name='OCR')
        self.rpc_support = False  # TODO: test

    def init_model(self):
        import easyocr
        if len(self.arg) > 0 or len(self.kwargs.keys()) > 0:
            self.reader = easyocr.Reader(*self.arg, **self.kwargs)
        else:
            self.reader = easyocr.Reader(['en', 'ch_sim'], quantize=False)

    def predict(self, img, auto_proc=False):
        if self.use_rpc():
            return self.rpc_call(img)
        self.wait_model_init()
        img = img_2_ndarray_rgb(img)
        # cv2.imwrite(gen_test_pic('ocr_input.bmp'), img)
        if auto_proc:
            # 灰度化
            img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
            # cv2.imwrite(gen_test_pic('ocr_gray.bmp'), img)

            # 二值化
            _, img = cv2.threshold(img, 127, 255, cv2.THRESH_BINARY)
            # cv2.imwrite(gen_test_pic('ocr_binary.bmp'), img)

            # 去噪
            img = cv2.GaussianBlur(img, (5, 5), 0)
            # cv2.imwrite(gen_test_pic('ocr_blur.bmp'), img)

            # 增强对比度
            img = cv2.equalizeHist(img)
            # cv2.imwrite(gen_test_pic('ocr_equal.bmp'), img)

        result = self.reader.readtext(img)
        ans = []
        for detection in result:
            text = detection[1]
            ans.append(text)
        return ans

    def rpc_req_encode(self, *args, **kwargs):
        args_list = list(args)
        args_list[0] = img_2_list(args_list[0])
        args = tuple(args_list)
        return args, kwargs

    def rpc_req_decode(self, *args, **kwargs):
        args_list = list(args)
        img = list_2_ndarray(args_list[0])
        if img.dtype != np.uint8:
            img = img.astype(np.uint8)
        args_list[0] = img
        args = tuple(args_list)
        return args, kwargs


if __name__ == '__main__':
    pass
    supper_res = SuperRes()
    scaled = supper_res.mock_rpc_call(get_test_pic('level.bmp'), scale_factor=3)
    ocr = OCR()
    ans = ocr.predict(scaled, debug=True)
    print(ans)
    ans = ocr.predict(scaled, debug=False)
    print(ans)
    # ans = ocr.predict(get_test_pic('attack.bmp'))
    # print(ans)



import time

from auto_easy.base.find_pic import ConfBase
from auto_easy.base.find_pic.model import PicV2, PicDetV2, PicDetConf, DetBox, nms_boxes, MPicDetV2
from auto_easy.base.find_pic.pic_handler import *
from auto_easy.base.image.draw import show_image_table
from auto_easy.base.image.process import contain_color
from auto_easy.constant import get_test_pic
from auto_easy.utils import concurrent_exec_one_func, logger


# 内部全部用cv2进行处理
def find_pic(source, search: PicV2, det_conf: PicDetConf = None) -> PicDetV2:
    pic_source = PicV2.new_auto(source)
    pic_search = PicV2.new_auto(search)
    conf: PicDetConf = ConfBase.new_conf_by_pry([PicDetConf(), pic_search.det_conf, det_conf])

    # print(det_conf.box)
    # 图片缩放匹配,这里并发进行匹配
    # TODO: 优化,记录历史匹配成功的scale,后续优先采纳
    if conf.range_scale is not None and conf.cur_scale is None:
        start = conf.range_scale[0]
        end = conf.range_scale[1]
        step = conf.range_scale[2]

        # 并发执行不同比例图片的匹配
        def _f(_conf):
            return find_pic(pic_source, pic_search, _conf)

        args = []
        for scale in np.arange(start, end, step):
            new_conf = conf.deepcopy()
            new_conf.range_scale = None
            new_conf.cur_scale = scale
            args.append(new_conf)
        if len(args) == 1:
            raise Exception('invalid range_scale, {}'.format(conf.range_scale))

        det_list = concurrent_exec_one_func(_f, args)
        # 合并结果
        first_det = det_list[0]
        first_det.merge_others(det_list[1:], dedup=True)
        return first_det

    cur_scale = conf.cur_scale if conf.cur_scale else 1

    # todo: 预留动态处理函数
    img_source = pic_source.cv2_img_bgr.copy()
    img_search = pic_search.cv2_img_bgr.copy()

    handlers = []

    # 根据参数加工图片
    if cur_scale != 1:
        handlers.append(PicResize(conf.cur_scale))

    if conf.color:
        handlers.append(PicColorKeep(conf.color))

    if not conf.rgb:
        handlers.append(PicColorGray())

    if conf.bg_remove:
        handlers.append(PicColorCanny())

    # 匹配区域, None/Box(0,0,0,0) 默认全屏
    match_box = conf.scaled_box if conf.box else None
    if match_box and not match_box.is_empty():
        handlers.append(PicCrop(match_box))  # 注: 这里是使用缩放后的区域进行裁剪

    his_source = []
    his_search = []
    titles = []
    for handler in handlers:
        img_source, img_search = handler.process(img_source, img_search)
        his_source.append(img_source)
        his_search.append(img_search)
        titles.append(handler.name)
        if conf.debug:
            logger.debug(f'img process: [{handler.name}]')

    if conf.debug:
        # print(len(his_source))
        show_image_table([his_source, his_search], titles)

    method = conf.method
    if conf.rgb:
        s_bgr = cv2.split(img_source)  # Blue Green Red
        i_bgr = cv2.split(img_search)
        weight = (0.33, 0.33, 0.33)
        resbgr = [0, 0, 0]
        for i in range(3):  # bgr
            resbgr[i] = cv2.matchTemplate(i_bgr[i], s_bgr[i], method)
        res = resbgr[0] * weight[0] + resbgr[1] * weight[1] + resbgr[2] * weight[2]
    else:
        res = cv2.matchTemplate(img_source, img_search, method)

    w, h = img_search.shape[1], img_search.shape[0]
    yloc, xloc = np.where(res >= conf.sim)
    # 存储匹配结果
    det_boxes = []
    for (x, y) in zip(xloc, yloc):
        similarity = res[y, x]

        # 如果是区域匹配, 则变换为原图的坐标
        actual_x = x if match_box is None else x + match_box.x1
        actual_y = y if match_box is None else y + match_box.y1
        x1, y1, x2, y2 = actual_x, actual_y, actual_x + w, actual_y + h
        sim = float(format(similarity, ".2f"))
        # 将框添加到列表
        det_box = DetBox(
            sim,
            x1, y1, x2, y2
        )

        if len(conf.contain_color) > 0:
            crop_img = pic_source.cv2_img_bgr[y1:y2, x1:x2]
            cover_rate = contain_color(crop_img, conf.contain_color)
            if cover_rate <= 0.1:
                continue

        det_box.name = f'{pic_search.name}-{sim}'
        det_boxes.append(det_box)

    # 执行非极大值抑制, 主要用于消除检测算法输出中的多余重叠框
    det_boxes = nms_boxes(det_boxes)

    sorted_boxes = sorted(det_boxes, key=lambda box: box.sim, reverse=True)
    if not conf.multi_match:
        sorted_boxes = sorted_boxes[:1]

    pic_det = PicDetV2(
        pic_search,
        sorted_boxes,
        cur_scale
    )
    if conf.debug:
        pic_source.show(boxes=sorted_boxes)
    return pic_det


def find_pics_v2(source,
                 searches,
                 pic_det_conf: PicDetConf = None,
                 ) -> MPicDetV2:
    start = time.time()
    pic_source = PicV2.new_auto(source)
    pics_search = [PicV2.new_auto(search) for search in searches]
    if len(pics_search) == 0:
        return MPicDetV2()

    def _det_one(_pic_search):
        return find_pic(pic_source, _pic_search, pic_det_conf)

    det_list = concurrent_exec_one_func(_det_one, pics_search)
    mdet = MPicDetV2(det_list)
    return mdet


if __name__ == '__main__':
    conf = PicDetConf()
    conf.rgb = False
    # conf.bg_remove = True
    conf.sim = 0.9
    # conf.debug = True
    # conf.box = '778,381,812,416'
    conf.multi_match = False
    # conf.range_scale = '0.90,1.05,0.01'
    conf.cur_scale = 0.91
    det = find_pic(
        get_test_pic('debug/fsdsyc1.bmp'),
        get_test_pic('debug/40_破坏者碉堡轰炸$$$level=40&type=主动技能.bmp'),
        conf
    )
    print(det)

    # conf = default_det_conf.deepcopy()
    # conf.rgb = False
    # # conf.bg_remove = True
    # conf.sim = 0.8
    # # conf.debug = True
    # conf.multi_match = True
    # conf.range_scale = '0.99,1.01,0.01'
    #
    # det = find_pic(
    #     get_test_pic('find_pic/source_pic_2.bmp'),
    #     get_test_pic('find_pic/search_pic_2.bmp'),
    #     conf
    # )
    # print(det)
    # mdet = find_pics_v2(
    #     get_test_pic('find_pic/pics_source_1.jpg'),
    #     [
    #         get_test_pic('find_pic/pics_search_1_1.bmp'),
    #         get_test_pic('find_pic/pics_search_1_1.bmp'),
    #         PicV2.new_auto(get_test_pic('find_pic/pics_search_1_2.png'), name='search1_2'),
    #         PicV2.new_auto(get_test_pic('find_pic/pics_search_1_2$$$sim=0.7.bmp'), name='search1_2'),
    #     ]
    # )
    # print(mdet)
    # search1 = PicV2.new_auto(get_test_pic('find_pic/pics_search_1_1.bmp'), name='search')
    # search2 = PicV2.new_auto(get_test_pic('find_pic/pics_search_1_2.png'), name='search')
    # mdet = find_pics_v2(
    #     get_test_pic('find_pic/pics_source_1.jpg'),
    #     [
    #         search1,
    #         search2,
    #     ],
    #     pics_det_conf=mdet_conf,
    # )
    # print(mdet)



import logging
import os

from logging.handlers import TimedRotatingFileHandler

logger_name = 'auto_easy'


def get_logger():
    return logging.getLogger(logger_name)


def get_log_formatter():
    # 定义日志格式
    datefmt = '%Y/%m/%d-%H:%M:%S'
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(filename)s:%(lineno)d - %(message)s', )
    formatter.default_time_format = datefmt
    formatter.default_msec_format = '%s.%03d'
    return formatter


def set_log_2_console(log_level=logging.DEBUG):
    console_handler = logging.StreamHandler()
    console_handler.setLevel(log_level)
    console_handler.setFormatter(get_log_formatter())
    get_logger().setLevel(log_level)
    get_logger().addHandler(console_handler)


def set_log_2_file(log_dir, log_level=logging.INFO):
    if not os.path.exists(log_dir):
        os.makedirs(log_dir)
    get_logger().setLevel(log_level)
    log_file = os.path.join(log_dir, f"auto_easy.log")
    logger = get_logger()
    file_handler = TimedRotatingFileHandler(log_file, when="D", interval=1, backupCount=3, encoding="utf-8")
    file_handler.setLevel(log_level)
    logger.addHandler(file_handler)


logger = get_logger()



import math
import random
import re
import sys
from enum import Enum
from typing import List


class Direction(Enum):
    UP = "UP"
    DOWN = "DOWN"
    LEFT = "LEFT"
    RIGHT = "RIGHT"


class Point:
    def __init__(self, x, y, name=None):
        self.name = name if name is not None else ''
        self._x = 0
        self._y = 0
        self.x = x
        self.y = y

    @property
    def x(self):
        return self._x

    @x.setter
    def x(self, value):
        self._x = int(value)

    @property
    def y(self):
        return self._y

    @y.setter
    def y(self, value):
        self._y = int(value)

    def __str__(self):
        return "{}<{},{}>".format(self.name, self.x, self.y)

    def is_nearby(self, p, x_radius=30, y_radius=30):
        return abs(self.x - p.x) <= x_radius and abs(self.y - p.y) <= y_radius

    def get_distance(self, p):
        dx = self.x - p.x
        dy = self.y - p.y
        return int(math.sqrt(dx ** 2 + dy ** 2))

    def calculate_angle(self, p):
        dx = p.x - self.x
        dy = p.y - self.y
        angle_rad = math.atan2(dy, dx)
        angle_deg = math.degrees(angle_rad)
        if angle_deg < 0:
            return angle_deg + 360
        else:
            return angle_deg

    def on_other_left(self, p):
        return self.x < p.x

    def on_other_right(self, p):
        return not self.on_other_left(p)

    def __sub__(self, other):
        return Point(self.x - other.x, self.y - other.y)

    def gen_point(self, x, y, max_x=-1, max_y=-1):
        p = Point(max(self.x + x, 0), max(self.y + y, 0))
        if max_x > 0:
            p.x = min(p.x, max_x)
        if max_y > 0:
            p.y = min(p.y, max_y)
        return p

    def move_x(self, dis, min_x=0, max_x=-1):
        self.x += dis
        if min_x >= 0:
            self.x = max(self.x, min_x)
        if max_x > 0:
            self.x = min(self.x, max_x)

    def tuple(self):
        return (self.x, self.y)


def check_distance(points: List[Point], user_max_x: float, user_max_y: float):
    """
    判断给定数组内所有元素的最大x轴距离和最大y轴距离是否分别小于max_x和max_y

    :param points: 包含Point对象的列表，代表坐标点集合
    :param max_x: x轴距离的最大值限制
    :param max_y: y轴距离的最大值限制
    :return: 如果满足条件返回True，否则返回False
    """
    if not points:
        return True
    min_x = max_x = points[0].x
    min_y = max_y = points[0].y
    for point in points[1:]:
        min_x = min(min_x, point.x)
        max_x = max(max_x, point.x)
        min_y = min(min_y, point.y)
        max_y = max(max_y, point.y)
    return (max_x - min_x) < user_max_x and (max_y - min_y) < user_max_y


def points_2_str(points) -> str:
    ls = [point.__str__() for point in points]

    return '|'.join(ls)


def group_points(points, dis_err) -> str:
    # 按照横坐标x进行升序排序（从左往右），如果x相同则按照纵坐标y升序排序（从上往下）
    sorted_points = sorted(points, key=lambda p: (p.y, p.x))

    # 用于存储分组后的结果（同一行或同一列的点为一组）
    groups = []
    current_group = []

    for i in range(len(sorted_points)):
        current_group.append(sorted_points[i])
        if i < len(sorted_points) - 1:
            next_point = sorted_points[i + 1]
            # 判断是否在同一列（横坐标距离在dis_err以内）
            if abs(next_point.y - current_group[0].y) <= dis_err:
                continue
            else:
                # 如果既不在同一行也不在同一列，当前组结束，添加到分组结果中并开启新组
                current_group = sorted(current_group, key=lambda p: p.x)
                groups.append(current_group)
                current_group = []
    # 添加最后一组（循环结束后最后一组还未添加）
    if current_group:
        current_group = sorted(current_group, key=lambda p: p.x)
        groups.append(current_group)
    return groups


def sort_and_group_points(points: List[Point], dis_err):
    groups = group_points(points, dis_err)
    merged_list = [element for sublist in groups for element in sublist]
    return merged_list


class Box:
    def __init__(self, tl_x, tl_y, rd_x, rd_y, name=""):
        tl_x, tl_y, rd_x, rd_y = int(tl_x), int(tl_y), int(rd_x), int(rd_y)
        if tl_x > rd_x or tl_y > rd_y:
            raise Exception(
                f"配置有问题，边框左上角和右下角信息错误, {tl_x}, {tl_y}, {rd_x}, {rd_y}, {name}; {tl_x > rd_x},{tl_y > rd_y}")
        self.name = name
        self.tl_point = Point(tl_x, tl_y, 'left_top')
        self.rd_point = Point(rd_x, rd_y, 'right_low')

    # 生成点，已左上角为起点，通过宽度和高度的比例
    def get_inner_point(self, width_rate=0.5, height_rate=0.5) -> Point:
        x = self.tl_point.x + (self.rd_point.x - self.tl_point.x) * width_rate
        y = self.tl_point.y + (self.rd_point.y - self.tl_point.y) * height_rate
        return Point(x, y)

    def get_mid_point(self, offset_rate=0) -> Point:
        w_rand = random.uniform(-offset_rate, offset_rate)
        h_rand = random.uniform(-offset_rate, offset_rate)
        return self.get_inner_point(0.5 + w_rand, 0.5 + h_rand)

    def get_rand_point(self):
        return self.get_inner_point(
            width_rate=random.uniform(0, 1),
            height_rate=random.uniform(0, 1),
        )

    def get_dis_to_center(self) -> int:
        mid_point = self.get_mid_point()
        return self.tl_point.get_distance(mid_point)

    def is_overlap(self, other) -> bool:
        """
        判断当前 Box 是否与另一 Box 重叠。
        :param other: 另一个 Box 对象
        :return: 如果重叠返回 True，否则返回 False
        """
        # 不重叠条件
        # A的右边在B的左边的左侧。
        # A的左边在B的右边的右侧。
        # A的底边在B的顶部的上方。
        # A的顶部在B的底边的下方
        # 检查是否不重叠
        if (self.x1 >= other.x2 or self.x2 <= other.x1 or
                self.y1 >= other.y2 or self.y2 <= other.y1):
            return False  # 不重叠

        return True  # 重叠

    def move(self, x, y):
        self.tl_point.x += x
        self.tl_point.y += y
        self.rd_point.x += x
        self.rd_point.y += y
        self.fix_overflow()

    def copy(self, x=0, y=0):
        new_box = Box(
            self.tl_point.x + x,
            self.tl_point.y + y,
            self.rd_point.x + x,
            self.rd_point.y + y,
            '')
        new_box.fix_overflow()
        return new_box

    def crop_by_rate(self, x1_r, y1_r, x2_r, y2_r):
        new_box = self.copy()
        new_box.tl_point.x += self.width * x1_r
        new_box.tl_point.y += self.height * y1_r
        new_box.rd_point.x -= self.width * (1 - x2_r)
        new_box.rd_point.y -= self.height * (1 - y2_r)
        new_box.fix_overflow()
        return new_box

    def copy_by_scale(self, scale):
        new_box = self.copy()
        new_box.tl_point.x += self.width * (1 - scale)
        new_box.tl_point.y += self.height * (1 - scale)
        new_box.rd_point.x -= self.width * (1 - scale)
        new_box.rd_point.y -= self.height * (1 - scale)
        new_box.fix_overflow()
        return new_box

    def fix_overflow(self, max_y=sys.maxsize, max_x=sys.maxsize):
        def _fix(cur_val, max_val):
            if cur_val < 0:
                return 0
            if cur_val > max_val:
                return max_val
            return cur_val

        self.tl_point.x = _fix(self.tl_point.x, max_x)
        self.tl_point.y = _fix(self.tl_point.y, max_y)
        self.rd_point.x = _fix(self.rd_point.x, max_x)
        self.rd_point.y = _fix(self.rd_point.y, max_y)

    def get_mid_dis(self, other):
        p1 = self.get_mid_point()
        p2 = other.get_mid_point()
        return p1.get_distance(p2)

    @property
    def x1(self):
        return self.tl_point.x

    @property
    def y1(self):
        return self.tl_point.y

    @property
    def x2(self):
        return self.rd_point.x

    @property
    def y2(self):
        return self.rd_point.y

    @property
    def area(self):
        return self.width * self.height

    @property
    def width(self):
        return self.x2 - self.x1

    @property
    def height(self):
        return self.y2 - self.y1

    def tuple(self):
        return (self.x1, self.y1, self.x2, self.y2)

    def is_empty(self):
        return self.width == 0 and self.height == 0

    @staticmethod
    def new_boxes_by_str(s):
        # 使用正则表达式提取<x1,y1,x2,y2>
        pattern = r'<(\d+),(\d+),(\d+),(\d+)>'
        matches = re.findall(pattern, s)
        # 打印提取的坐标
        boxes = []
        for match in matches:
            x1, y1, x2, y2 = match
            boxes.append(Box(x1, y1, x2, y2))
        return boxes

    def __str__(self):
        return "<{},{},{},{}>".format(self.tl_point.x, self.tl_point.y, self.rd_point.x, self.rd_point.y)

    def __eq__(self, other):
        if isinstance(other, Box):
            # 都为None的情况
            if self is None and other is None:
                return True
            # 均不为None的情况，比较坐标是否一致
            elif self is not None and other is not None:
                return self.x1 == other.x1 and self.y1 == other.y1 and self.x2 == other.x2 and self.y2 == other.y2
            else:
                return False
        return False


def sort_boxes_by_group(boxes: list[Box], dis_err=20):
    # 从左到右, 从上往下排序
    points = [box.tl_point for box in boxes]
    sorted_points = sort_and_group_points(points, dis_err=dis_err)
    # points的顺序映射到box的顺序
    sort_boxes = []
    for point in sorted_points:
        for box in boxes:
            if id(box.tl_point) == id(point):
                sort_boxes.append(box)
                break
    return sort_boxes


def group_box(boxes: list[Box], dis_err=20):
    points = [box.tl_point for box in boxes]
    groups = group_points(points, dis_err=dis_err)
    box_matrix = []
    for row in groups:
        box_row = []
        for point in row:
            for box in boxes:
                if id(box.tl_point) == id(point):
                    box_row.append(box)
                    break
        box_matrix.append(box_row)
    return box_matrix


class Circle:
    def __init__(self, name, point: Point, radius: int):
        self.name = name
        self.point = point
        self.radius = radius

    def get_rand_point_by_dis(self) -> Point:
        # 越靠近中心的概率越高；利用了圆的极坐标方程
        r = self.radius * math.sqrt(random.random())
        theta = 2 * math.pi * random.random()
        dx = r * math.cos(theta)
        dy = r * math.sin(theta)
        return Point(self.point.x + dx, self.point.y + dy, '{}内随机点'.format(self.name))

    def contain_point(self, point: Point) -> bool:
        dis = self.point.get_distance(point)
        return dis < self.radius

    def in_hr_line(self, p1, p2) -> bool:
        if p1.x < self.point.x < p2.x:
            return True
        if p2.x < self.point.x < p1.x:
            return True
        return False

    def touch_line(self, p1, p2):
        # 计算圆心到直线的距离
        numerator = abs((p2.x - p1.x) * (p1.y - self.point.y) - (p1.x - self.point.x) * (p2.y - p1.y))
        denominator = math.sqrt((p2.x - p1.x) ** 2 + (p2.y - p1.y) ** 2)
        if denominator == 0:
            distance = math.sqrt((p1.x - self.point.x) ** 2 + (p1.y - self.point.y) ** 2)
        else:
            distance = numerator / denominator
        # 判断距离和半径的关系
        if distance <= self.radius:
            return True
        return False


if __name__ == '__main__':
    box = Box(84, 261, 372, 418)
    # print(box.get_dis_to_center())
    cc = Circle('', Point(753, 298), 90)
    print(cc.touch_line(Point(466, 306), Point(914, 357)))
    # 人物(<466,306>), 终点(<914,357>), 门柱(<753,298>)



from auto_easy.core.core import *
from auto_easy.core.win_ai import *
from auto_easy.core.win_find_color import *
from auto_easy.core.win_find_pic import *
from auto_easy.core.win_monitor import *
from auto_easy.core.win_ocr import *
from auto_easy.core.win_show import *
from auto_easy.core.win_supper_res import *



import json
import time
import warnings
from typing import List

import numpy as np
import torch

from auto_easy.base.ai.model_mgr_v2 import AIModelBase
from auto_easy.base.image.cvt import img_2_ndarray_rgb, img_2_list, list_2_ndarray
from auto_easy.models import MYoloItem, YoloItem, Box
from auto_easy.utils import logger, cls_to_dict, set_obj_by_dict

warnings.filterwarnings("ignore", category=FutureWarning)


class ModelConf:
    def __init__(self, name, pt_path, prob=0.5, iou=0.45):
        self.name = name
        self.pt_path = pt_path
        self.prob = prob
        self.iou = iou
        self.agnostic = False
        self.multi_label = True


class YoloObj:
    def __init__(self):
        self.id = -1
        self.name = ''
        self.score = 0
        self.x1 = 0
        self.y1 = 0
        self.x2 = 0
        self.y2 = 0

    def __str__(self):
        return f'{self.id}({self.name}):<{self.x1}, {self.y1}, {self.x2}, {self.y2}> {self.score}'


class YoloObjs:
    def __init__(self, conf: ModelConf, dets: List[YoloObj]):
        self.objs = dets
        self.conf = conf
        self.screen_box = None

        self.result_str = ''  # TODO: 待删除
        self.create_ts = time.time()

    def to_dict(self):
        return {}

    def __str__(self):
        return '|'.join([str(det) for det in self.objs])

    @staticmethod
    def new_obj_by_str(s):
        segments = s.split('|')
        dets = []
        for segment in segments:
            name, coordinates = segment.split('<')
            x, y = coordinates.rstrip('>').split(',')
            obj = YoloObj()
            obj.name = name
            obj.x1 = int(x)
            obj.y1 = int(y)
            obj.x2 = int(x)
            obj.y2 = int(y)
            dets.append(obj)
        return YoloObjs(ModelConf('', ''), dets)


class AIYolo(AIModelBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class AIYoloV5(AIYolo):
    def __init__(self, conf: ModelConf):
        self.conf = conf
        self.model = None
        super().__init__(self.conf.name)

    def init_model(self):
        conf = self.conf
        # # 禁用所有级别的日志记录
        # logger.disabled = True
        model = torch.hub.load('ultralytics/yolov5', 'custom', verbose=False, path=conf.pt_path)

        # 将模型设置为eval模式，关闭一些训练相关的模块（如Dropout等），提升性能
        model.eval()
        # 设置模型相关参数（可选操作）
        model.det_conf = conf.prob  # 设置置信度阈值，只有置信度高于此值的检测结果才会被保留
        model.iou = conf.iou  # 设置非极大值抑制（NMS）的交并比（IoU）阈值
        model.agnostic = conf.agnostic  # 设置是否进行类别无关的NMS，False表示按照类别进行NMS筛选
        model.multi_label = conf.multi_label  # 设置是否允许每个检测框有多个标签，False表示每个检测框只对应一个类别
        self.model = model
        logger.debug(f'模型({conf.name})完成加载')

    def predict(self, img) -> MYoloItem:
        if self.use_rpc():
            return self.rpc_call(img)
        self.wait_model_init()
        img = img_2_ndarray_rgb(img)
        height, width = img.shape[:2]
        with torch.no_grad():
            # 使用模型进行预测，在这个上下文环境中，不会进行梯度计算，节省计算资源，提升性能
            results = self.model(img)
            # 处理预测结果，整理成要求的格式
            detections = results.pandas().xyxy[0].to_dict(orient='records')
            items = []
            for idx, detection in enumerate(detections, start=1):
                name = detection['name']
                score = round(detection['confidence'], 2)
                x1 = int(detection['xmin'])
                y1 = int(detection['ymin'])
                x2 = int(detection['xmax'])
                y2 = int(detection['ymax'])
                item = YoloItem(name, score, Box(x1, y1, x2, y2), src_box=Box(0, 0, height, width))
                items.append(item)
        return MYoloItem(items, self.conf.name)

    def rpc_req_encode(self, *args, **kwargs):
        if len(args) != 1:
            raise Exception('args length error')
        args_list = list(args)
        args_list[0] = img_2_list(args_list[0])
        args = tuple(args_list)
        return args, kwargs

    def rpc_req_decode(self, *args, **kwargs):
        args_list = list(args)
        img = list_2_ndarray(args_list[0])
        if img.dtype != np.uint8:
            img = img.astype(np.uint8)
        args_list[0] = img
        args = tuple(args_list)
        return args, kwargs

    def rpc_resp_encode(self, origin_resp: MYoloItem):
        return json.dumps(cls_to_dict(origin_resp))

    def rpc_resp_decode(self, str_resp):
        d = json.loads(str_resp)
        items = []
        for v in d['items']:
            obj = YoloItem(
                v['name'],
                v['score'],
                match_box=Box(
                    v['box']['tl_point']['_x'], v['box']['tl_point']['_y'],
                    v['box']['rd_point']['_x'], v['box']['rd_point']['_y'],
                ),
                src_box=Box(
                    v['src_box']['tl_point']['_x'], v['src_box']['tl_point']['_y'],
                    v['src_box']['rd_point']['_x'], v['src_box']['rd_point']['_y'],
                ),
            )
            items.append(obj)

        res = MYoloItem([], d.get('model_name', ''))
        set_obj_by_dict(res, d)
        res.items = items
        return res


if __name__ == '__main__':
    pass



import threading
import time

from auto_easy.base.windows import Window
from auto_easy.utils import logger


class TimerProcessor:
    def __init__(self, interval, f):
        self.interval = interval
        self.f = f

    def _timer_call(self):
        interval_ms = self.interval * 1000
        while True:
            now_ms = time.time() * 1000
            next_call = now_ms + (interval_ms - (now_ms % interval_ms))
            delay_ms = next_call - now_ms
            time.sleep(delay_ms / 1000)
            self.f()

    def async_run(self):
        th = threading.Thread(target=self._timer_call)
        th.daemon = True
        th.start()


class WinTimer(Window):
    def __init__(self, window_id=None):
        Window.__init__(self, window_id=window_id)

    def add_timer_minitor(self, interval, f):
        def _f():
            f(self)

        logger.debug('add_timer_minitor, interval=%f, f=%s', interval, f)
        TimerProcessor(interval, _f).async_run()


if __name__ == '__main__':
    win = WinTimer('Phone-VB')

    win.add_timer_minitor(1.5, lambda x: print(time.time()))

    time.sleep(10)



import logging
from unittest import TestCase

from auto_easy import DAG, ExecutorDebug, Ctx, DAGLayerSimple
from auto_easy.global_log import set_log_2_console, logger


class TestDag(DAG):
    def __init__(self):
        super().__init__('测试重试功能')

    def init(self):
        self.add_layer(ExecutorDebug('layer1'))
        self.add_layer(ExecutorDebug('layer2', exec_wait=1))
        self.add_layer(ExecutorDebug('layer3', hit_ret=True))


class TestDAG(TestCase):
    def test_dag(self):
        dag = TestDag()
        ok = dag.run(ctx=Ctx())
        self.assertTrue(ok)
        # logger.debug(ok)

    def test_subdag(self):
        set_log_2_console(logging.INFO)
        logger.debug('test_subdag')
        class TestSubDag(DAG):
            def __init__(self):
                super().__init__('测试Sub功能')

            def init(self):
                self.add_layer(TestDag())
                self.add_layer(DAGLayerSimple(ExecutorDebug('after_sub_dag',hit_ret=True)))

        dag2 = TestSubDag()
        ok = dag2.run(ctx=Ctx())
        self.assertTrue(ok)



import time
from abc import abstractmethod
from typing import List

from auto_easy.dag.executor import Executor
from auto_easy.models import Ctx
from auto_easy.utils import set_obj_by_dict, Timeout, logger, concurrent_exec_multi_func_one_arg


class LayerConf:
    def __init__(self, skip_err=False):
        # 通用配置: 控制DAG图的运行逻辑
        self.name = ''
        self.skip_err = skip_err  # 忽略该层错误,继续往下执行,达到类似节点跳过的效果

        # 定制化配置
        self.type = DAGLayerSimple

    @staticmethod
    def new_by_json(js):
        conf = LayerConf()
        set_obj_by_dict(conf, js)
        return conf


class DAGLayerDef(Executor):
    def __init__(self, name='', conf: LayerConf = None):
        super().__init__(name)
        self.conf = conf if conf is not None else LayerConf()

    @abstractmethod
    def hit(self, ctx: Ctx) -> bool:
        pass

    @abstractmethod
    def exec(self, ctx: Ctx):
        pass


class DAGLayerSimple(DAGLayerDef):
    def __init__(self, executor: Executor, conf: LayerConf = None):
        super().__init__(executor.name, conf)
        self.executor = executor

    def hit(self, ctx: Ctx) -> bool:
        return self.executor.hit(ctx)

    def exec(self, ctx: Ctx):
        return self.executor.exec(ctx)


class LayerLoopConf(LayerConf):
    def __init__(self):
        super().__init__()
        self.loop_times = 1  # 最大循环次数
        self.loop_suc_times = 0  # 至少成功的次数


class DAGLayerLoop(DAGLayerSimple):
    def __init__(self, executor: Executor, conf: LayerLoopConf = None):
        conf = conf if conf is not None else LayerLoopConf()
        super().__init__(executor, conf)
        self.conf = conf

    def exec(self, ctx: Ctx):
        success_cnt = 0
        for i in range(1, self.conf.loop_times):
            success = self.executor.run(ctx)
            if success:
                success_cnt += 1

            if not success:
                return success_cnt >= self.conf.loop_suc_times
        return True


class DAGLayerSwitch(DAGLayerDef):
    def __init__(self, executors: List[Executor], switch_to=1):
        super().__init__('多分支互斥路由')
        self.executors = executors
        self.switch_to = switch_to
        self.target_exec = None

    def hit(self, ctx: Ctx) -> bool:
        # 实现switch效果,选择命中的exec
        to = Timeout(self.switch_to)
        self.target_exec = None
        while to.not_timeout():
            for executor in self.executors:
                if executor.hit(ctx):
                    self.target_exec = executor
                    break
            if self.target_exec:
                break
            time.sleep(self.switch_to / 10)
        if self.target_exec:
            return True
        return False

    def exec(self, ctx: Ctx):
        return self.target_exec.run(ctx)


class SwitchBranch:
    def __init__(self, executor: Executor, is_finish=False):
        self.executor: Executor = executor
        self.is_finish = is_finish


class DAGLayerLoopSwitch(DAGLayerDef):
    def __init__(self, loop_to=1, min_loop_time=0, loop_sleep=0.1, bf_sleep=0):
        super().__init__('循环多分支路由')
        self.branches: list[SwitchBranch] = []
        self.loop_to = loop_to
        self.loop_sleep = loop_sleep
        self.min_loop_time = min_loop_time
        self.bf_sleep = bf_sleep

    def hit(self, ctx: Ctx) -> bool:
        return True

    def add_branch(self, executor: Executor, is_finish=False):
        self.branches.append(SwitchBranch(executor, is_finish))

    def exec(self, ctx: Ctx):
        time.sleep(self.bf_sleep)
        to = Timeout(self.loop_to)
        while to.not_timeout():
            target_branch = None
            funcs = [branch.executor.hit for branch in self.branches]
            hits = concurrent_exec_multi_func_one_arg(funcs, ctx)
            for i, is_hit in enumerate(hits):
                if is_hit:
                    target_branch = self.branches[i]
                    break
            if not target_branch:
                time.sleep(self.loop_sleep)
                continue
            ok = target_branch.executor.run(ctx)
            if not ok:
                logger.debug('Layer(循环多分支路由) 执行失败, Executor: {}'.format(target_branch.executor.name))
                return False

            if target_branch.is_finish:
                return True
            time.sleep(self.loop_sleep)
        logger.error('Layer(循环多分支路由) 执行失败, 超时: {}'.format(self.loop_to))
        return False


class DAGLayerSwitchOne(DAGLayerDef):
    def __init__(self, loop_to=1, loop_sleep=0.1, bf_sleep=0):
        super().__init__('多分支选一执行')
        self.branches: list[SwitchBranch] = []
        self.loop_to = loop_to
        self.loop_sleep = loop_sleep
        self.bf_sleep = bf_sleep

    def hit(self, ctx: Ctx) -> bool:
        return True

    def add_branch(self, executor: Executor, is_finish=False):
        self.branches.append(SwitchBranch(executor, is_finish))

    def exec(self, ctx: Ctx):
        time.sleep(self.bf_sleep)
        to = Timeout(self.loop_to)
        target_branch = None
        while to.not_timeout():
            funcs = [branch.executor.hit for branch in self.branches]
            hits = concurrent_exec_multi_func_one_arg(funcs, ctx)
            for i, is_hit in enumerate(hits):
                if is_hit:
                    target_branch = self.branches[i]
                    break
            if target_branch:
                break
            time.sleep(self.loop_sleep)

        if not target_branch:
            logger.error('Layer(多分支选一执行) 规定时间内未选中任意分支, 超时: {}, 分支名: {}'
                         .format(self.loop_to, [branch.executor.name for branch in self.branches]))
            return False

        ok = target_branch.executor.run(ctx)
        if not ok:
            logger.error('Layer(多分支选一执行) 执行分支({})失败'.format(target_branch.executor.name))
            return False
        return True


if __name__ == '__main__':
    pass



from setuptools import setup, find_packages

setup(
    name='auto_easy',  # 打包后的名称
    version='0.1.19',  # 版本号
    author='Jefung',
    author_email='jefung865424525@gmail.com',
    description='make automated script development easier',
    long_description=open('README.md', encoding='utf-8').read(),  # 从README.md中读取详细描述
    long_description_content_type='text/markdown',
    url='https://github.com/Jefung/auto_easy',  # 项目的URL
    packages=find_packages(),  # 自动发现所有的���
    classifiers=[  # 根据您的许可证调整
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.9',  # 最低Python版本
    include_package_data=True,  # 包含包中的数据文件
    # package_data 可以不必指定，MANIFEST.in 中的规则将处理
    package_data={
        'auto_easy': ['ai/superres/*', 'image/*'],  # 如果希望使用 package_data，而不使用 manifest
    },
)
